package cz.fidentis.analyst.engines.face.impl;

import cz.fidentis.analyst.data.face.HumanFace;
import cz.fidentis.analyst.data.kdtree.KdTree;
import cz.fidentis.analyst.data.landmarks.MeshVicinity;
import cz.fidentis.analyst.data.mesh.MeshFacet;
import cz.fidentis.analyst.data.mesh.MeshFactory;
import cz.fidentis.analyst.data.mesh.MeshModel;
import cz.fidentis.analyst.data.mesh.MeshPoint;
import cz.fidentis.analyst.data.octree.Octree;
import cz.fidentis.analyst.engines.bbox.BoundingBoxConfig;
import cz.fidentis.analyst.engines.bbox.BoundingBoxServices;
import cz.fidentis.analyst.engines.curvature.CurvatureConfig;
import cz.fidentis.analyst.engines.curvature.CurvatureServices;
import cz.fidentis.analyst.engines.distance.MeshDistanceConfig;
import cz.fidentis.analyst.engines.distance.MeshDistanceServices;
import cz.fidentis.analyst.engines.glyphs.GlyphServices;
import cz.fidentis.analyst.engines.glyphs.GlyphsConfig;
import cz.fidentis.analyst.engines.point2surface.PointToSurfaceDistanceConfig;
import cz.fidentis.analyst.engines.point2surface.PointToSurfaceDistanceServices;
import cz.fidentis.analyst.engines.sampling.PointSamplingConfig;
import cz.fidentis.analyst.engines.symmetry.SymmetryConfig;
import cz.fidentis.analyst.engines.symmetry.SymmetryServices;
import cz.fidentis.analyst.engines.face.FaceStateServices;
import cz.fidentis.analyst.engines.landmarks.LandmarkServices;

import java.util.Collection;

/**
 * Services for a single human face.
 *
 * @author Radek Oslejsek
 */
public class FaceStateServicesImpl implements FaceStateServices {

    /**
     * Manages a k-d tree storing the mesh of given human face.
     *
     * @param face Human face
     * @param mode Operation to be performed
     */
    public void updateKdTree(HumanFace face, Mode mode) {
        switch (mode) {
            case RECOMPUTE_IF_PRESENT -> {
                if (face.hasKdTree()) {
                    face.setKdTree(KdTree.create(face.getMeshModel()));
                }
            }
            case COMPUTE_IF_ABSENT -> {
                if (!face.hasKdTree()) {
                    face.setKdTree(KdTree.create(face.getMeshModel()));
                }
            }
            case COMPUTE_ALWAYS -> face.setKdTree(KdTree.create(face.getMeshModel()));
            case DELETE -> face.setKdTree(null);
            default -> throw new IllegalStateException("Unexpected value: " + mode);
        }
    }

    /**
     * Manages an octree storing the mesh of given human face.
     *
     * @param face Human face
     * @param mode Operation to be performed
     */
    public void updateOctree(HumanFace face, Mode mode) {
        switch (mode) {
            case RECOMPUTE_IF_PRESENT -> {
                if (face.hasOctree()) {
                    face.setOctree(Octree.create(face.getMeshModel()));
                }
            }
            case COMPUTE_IF_ABSENT -> {
                if (!face.hasOctree()) {
                    face.setOctree(Octree.create(face.getMeshModel()));
                }
            }
            case COMPUTE_ALWAYS -> face.setOctree(Octree.create(face.getMeshModel()));
            case DELETE -> face.setOctree(null);
            default -> throw new IllegalStateException("Unexpected value: " + mode);
        }
    }

    /**
     * Manages bounding box of the given human face.
     *
     * @param face Human face
     * @param mode Operation to be performed
     */
    public void updateBoundingBox(HumanFace face, Mode mode) {
        switch (mode) {
            case RECOMPUTE_IF_PRESENT -> {
                if (face.hasBoundingBox()) {
                    face.setBoundingBox(BoundingBoxServices.compute(face.getMeshModel(), new BoundingBoxConfig()));
                }
            }
            case COMPUTE_IF_ABSENT -> {
                if (!face.hasBoundingBox()) {
                    face.setBoundingBox(BoundingBoxServices.compute(face.getMeshModel(), new BoundingBoxConfig()));
                }
            }
            case COMPUTE_ALWAYS -> face.setBoundingBox(BoundingBoxServices.compute(face.getMeshModel(), new BoundingBoxConfig()));
            case DELETE -> face.setBoundingBox(null);
            default -> throw new IllegalStateException("Unexpected value: " + mode);
        }
    }

    /**
     * Manages glyphs of the given human face.
     *
     * @param face Human face
     * @param mode Operation to be performed
     * @param method Sub-sampling method
     * @param maxSamples Maximum number of glyphs
     */
    public void updateGlyphs(HumanFace face, Mode mode, PointSamplingConfig.Method method, int maxSamples) {
        var config = new GlyphsConfig(new PointSamplingConfig(method, maxSamples));

        switch (mode) {
            case RECOMPUTE_IF_PRESENT -> {
                if (face.hasGlyphs()) {
                    face.setGlyphs(GlyphServices.calculateGlyphs(face.getMeshModel(), config));
                }
            }
            case COMPUTE_IF_ABSENT -> {
                if (!face.hasGlyphs()) {
                    face.setGlyphs(GlyphServices.calculateGlyphs(face.getMeshModel(), config));
                }
            }
            case COMPUTE_ALWAYS -> face.setGlyphs(GlyphServices.calculateGlyphs(face.getMeshModel(), config));
            case DELETE -> face.setGlyphs(null);
            default -> throw new IllegalStateException("Unexpected value: " + mode);
        }
    }

    /**
     * Manages curvature values of the given human face.
     *
     * @param face Human face
     * @param mode Operation to be performed
     */
    public void updateCurvature(HumanFace face, Mode mode) {
        switch (mode) {
            case RECOMPUTE_IF_PRESENT -> {
                if (face.getMeshModel().hasCurvature()) {
                    CurvatureServices.computeAndSet(face.getMeshModel(), new CurvatureConfig());
                }
            }
            case COMPUTE_IF_ABSENT -> {
                if (!face.getMeshModel().hasCurvature()) {
                    CurvatureServices.computeAndSet(face.getMeshModel(), new CurvatureConfig());
                }
            }
            case COMPUTE_ALWAYS -> CurvatureServices.computeAndSet(face.getMeshModel(), new CurvatureConfig());
            case DELETE -> throw new UnsupportedOperationException();
            default -> throw new IllegalStateException("Unexpected value: " + mode);
        }
    }

    /**
     * Computes vicinity of all landmarks of the given human face.
     *
     * @param face Human face
     * @param mode Operation to be performed
     */
    public void updateLandmarksVicinity(HumanFace face, Mode mode) {
        switch (mode) {
            case RECOMPUTE_IF_PRESENT -> {
                if (face.getMeshModel().hasCurvature()) {
                    computeVicinity(face);
                }
            }
            case COMPUTE_IF_ABSENT -> {
                if (!face.getMeshModel().hasCurvature()) {
                    computeVicinity(face);
                }
            }
            case COMPUTE_ALWAYS -> computeVicinity(face);
            case DELETE -> face.getAllLandmarks().forEach(landmark -> landmark.setMeshVicinity(null));
            default -> throw new IllegalStateException("Unexpected value: " + mode);
        }
    }

    /**
     * Computes symmetry plane from mesh or landmarks.
     *
     * @param face Human face
     * @param mode Operation to be performed
     * @param meshSymmetryConfig Configuration of the mesh symmetry calculation (when needed).
     *                           If {@code null}, then the symmetry plane is computed from landmarks.
     */
    public void updateSymmetryPlane(HumanFace face, Mode mode, SymmetryConfig meshSymmetryConfig) {
        switch (mode) {
            case RECOMPUTE_IF_PRESENT -> {
                if (face.hasSymmetryPlane()) {
                    if (meshSymmetryConfig != null) {
                        face.setSymmetryPlane(SymmetryServices.estimate(face.getMeshModel(), meshSymmetryConfig));
                    } else {
                        face.setSymmetryPlane(LandmarkServices.computeSymmetryPlane(face.getAllLandmarks()));
                    }
                }
            }
            case COMPUTE_IF_ABSENT -> {
                if (!face.hasSymmetryPlane()) {
                    if (meshSymmetryConfig != null) {
                        face.setSymmetryPlane(SymmetryServices.estimate(face.getMeshModel(), meshSymmetryConfig));
                    } else {
                        face.setSymmetryPlane(LandmarkServices.computeSymmetryPlane(face.getAllLandmarks()));
                    }
                }
            }
            case COMPUTE_ALWAYS -> {
                if (meshSymmetryConfig != null) {
                    face.setSymmetryPlane(SymmetryServices.estimate(face.getMeshModel(), meshSymmetryConfig));
                } else {
                    face.setSymmetryPlane(LandmarkServices.computeSymmetryPlane(face.getAllLandmarks()));
                }
            }
            case DELETE -> face.setSymmetryPlane(null);
            default -> throw new IllegalStateException("Unexpected value: " + mode);
        }
    }

    /**
     * Calculation of precision of the symmetry plane.
     *
     * @param face Human face
     * @return precision of the symmetry. The closer to zero, the better. Returns -1 if no symmetry plane is available.
     */
    public double calculateSymmetryPlanePrecision(HumanFace face) {
        if (!face.hasSymmetryPlane()) {
            return -1.0;
        }

        MeshModel clone = MeshFactory.cloneMeshModel(face.getMeshModel());

        for (MeshFacet facet: clone.getFacets()) { // invert mesh
            facet.getVertices().parallelStream()
                    .forEach(v -> v.getPosition().set(face.getSymmetryPlane().reflectPointOverPlane(v.getPosition())));
        }

        FaceStateServices.updateKdTree(face, FaceStateServices.Mode.COMPUTE_IF_ABSENT);
        return MeshDistanceServices.measure(
                clone,
                new MeshDistanceConfig(
                        MeshDistanceConfig.Method.POINT_TO_POINT_NEAREST_NEIGHBORS,
                        face.getKdTree(),
                        false,
                        true)).getDistanceStats().getAverage();
    }

    protected void computeVicinity(HumanFace face) {
        updateKdTree(face, Mode.COMPUTE_IF_ABSENT);
        for (var landmark: face.getAllLandmarks()) {
            var config = new PointToSurfaceDistanceConfig(
                    PointToSurfaceDistanceConfig.Method.POINT_TO_VERTICES,
                    landmark.getPosition(),
                    true);
            var distInfo = PointToSurfaceDistanceServices.measure(face.getKdTree(), config);
            var vicinity = new MeshVicinity(
                    distInfo.getLeft(),
                    distInfo.getRight().values().stream()
                            .flatMap(Collection::stream)
                            .map(MeshPoint::getPosition)
                            .findAny()
                            .orElse(null));
            landmark.setMeshVicinity(vicinity);
        }
    }

}
