package cz.fidentis.analyst.gui.app.tools;

import com.jogamp.opengl.GLAutoDrawable;
import com.jogamp.opengl.GLCapabilities;
import com.jogamp.opengl.GLDrawableFactory;
import com.jogamp.opengl.GLProfile;
import cz.fidentis.analyst.data.mesh.MeshFacet;
import cz.fidentis.analyst.data.mesh.MeshIO;
import cz.fidentis.analyst.engines.distance.MeshDistanceConfig;
import cz.fidentis.analyst.engines.distance.MeshDistanceVisitor;
import cz.fidentis.analyst.engines.distance.measurement.DistanceRecord;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Class for testing the performance of GPU ray-casting vs CPU ray-casting.
 *
 * @author Pavol Kycina
 */
public class MeshDistanceGPUBenchmark {

    /**
     * Benchmark for testing acceleration of mesh distance computation on GPU vs CPU.
     * <p>
     * To run performance nvidia GPU on multi-GPU linux system, you can set these environmental variables:
     * __NV_PRIME_RENDER_OFFLOAD=1; __GLX_VENDOR_LIBRARY_NAME=nvidia
     * </p>
     * @param args paths to the .obj files, which will be used as secondary faces, first will be used as a main (primary)
     * @throws IOException when testing models cannot be loaded
     */
    public static void main(String[] args) throws IOException {
        System.out.println("*** Starting benchmark ***");

        List<MeshFacet> facetsToTest = new ArrayList<>();
        List<MeshFacet> mainFacets = loadModel(args[0]);

        for(String path : Arrays.stream(args).skip(1).toList()) {
            facetsToTest.addAll(loadModel(path));
        }

        final boolean relativeDistance = true;

        final int primaryIterations = 3;
        final int secondaryIterations = 2;

        long totalBuildGPU = 0;
        long totalBuildCPU = 0;
        long totalRaysGPU = 0;
        long totalRaysCPU = 0;

        System.out.println("There will be " + primaryIterations + " primary and " + secondaryIterations + " secondary iterations with " + facetsToTest.size() + " facets to test.");

        long startT = System.currentTimeMillis();

        GLProfile glProfile = GLProfile.getDefault();
        GLCapabilities glCapabilities = new GLCapabilities(glProfile);
        GLAutoDrawable drawable = GLDrawableFactory.getFactory(glProfile).createOffscreenAutoDrawable(null, glCapabilities, null, 1, 1);
        drawable.display(); // initialize context

        System.out.println("context create time: " + (System.currentTimeMillis() - startT));



        for (int i = 0; i < primaryIterations; i++) {
            long startTime = System.currentTimeMillis();
            MeshDistanceVisitor distVisitorGPU = new MeshDistanceConfig(MeshDistanceConfig.Method.RAY_CASTING_GPU, mainFacets, drawable.getContext(), relativeDistance, true).getVisitor();
            totalBuildGPU = System.currentTimeMillis() - startTime;

            startTime = System.currentTimeMillis();
            MeshDistanceVisitor distVisitorCPU = new MeshDistanceConfig(MeshDistanceConfig.Method.RAY_CASTING, mainFacets, null, relativeDistance, true).getVisitor();
            totalBuildCPU = System.currentTimeMillis() - startTime;

            for (int j = 0; j < secondaryIterations; j++) {
                long time = bulkVisit(distVisitorCPU, facetsToTest);
                totalRaysCPU += time;
                System.out.println("\nCPU time: " + time);

                time = bulkVisit(distVisitorGPU, facetsToTest);
                totalRaysGPU += time;
                System.out.println("GPU time: " + time);
            }

            for(MeshFacet facet : facetsToTest) {
                int matches = countMatches(distVisitorGPU, distVisitorCPU, facet);
                int raysCount = distVisitorGPU.getDistancesOfVisitedFacets().getFacetMeasurement(facet).size();
                float percent = (int)((1000f * matches) / raysCount) / 10f;
                System.out.println("result matches: " + matches + " / " + raysCount + "  " + percent + "%");
            }

            distVisitorGPU.dispose();
        }

        drawable.destroy();

        System.out.println("\nAveragare GPU algorithm structures build time: " + totalBuildGPU);
        System.out.println("Average CPU algorithm structures build time: " + totalBuildCPU);

        float averageCPU = (float) totalRaysCPU / (secondaryIterations * primaryIterations);
        float averageGPU = (float) totalRaysGPU / (secondaryIterations * primaryIterations);

        System.out.println("\n  *******************************  ");
        System.out.println("   average for CPU rays time: " + averageCPU + " ms");
        System.out.println("   average for GPU rays time: " + averageGPU + " ms");
        System.out.println("   speed-up rays time: " + averageCPU / averageGPU);
        System.out.println("   speed-up initialization time: " + (float) totalBuildCPU / totalBuildGPU);
        System.out.println("  *******************************  \n");
    }

    static List<MeshFacet> loadModel(String path) throws IOException {
        List<MeshFacet> facets =  MeshIO.readMeshModel(new File(path)).getFacets();

        int verts = facets.stream().map(x -> x.getVertices().size()).reduce(0, Integer::sum);
        int tri = facets.stream().map(x -> x.getTriangles().size()).reduce(0, Integer::sum);

        System.out.println("loaded " + facets.size() + " facets with total vertices: " + verts + " and triangles: " + tri);

        return facets;
    }

    static long bulkVisit(MeshDistanceVisitor visitor, List<MeshFacet> facets) {
        long startTime = System.currentTimeMillis();
        for(MeshFacet facet : facets) {
            visitor.visitMeshFacet(facet);
        }
        return System.currentTimeMillis() - startTime;
    }

    static int countMatches(MeshDistanceVisitor distVisitorGPU, MeshDistanceVisitor distVisitorCPU, MeshFacet facet) {
        int matches = 0;

        for (int i = 0; i < distVisitorGPU.getDistancesOfVisitedFacets().getFacetMeasurement(facet).size(); i++) {
            DistanceRecord recGPU = distVisitorGPU.getDistancesOfVisitedFacets().getFacetMeasurement(facet).get(i);
            DistanceRecord recCPU = distVisitorCPU.getDistancesOfVisitedFacets().getFacetMeasurement(facet).get(i);

            double dGPU = recGPU.getDistance();
            double dCPU = recCPU.getDistance();

            if((int)dGPU == (int)dCPU) {
                matches++;
            }else {
                //System.out.println("NOT MATCHING: (gpu) " + dGPU + " (cpu) " + dCPU + " i: " + i);
            }
        }
        return matches;
    }

}
