package cz.fidentis.analyst.gui.project;

import cz.fidentis.analyst.Logger;
import cz.fidentis.analyst.gui.elements.ProgressDialog;
import cz.fidentis.analyst.project.FacesProxy;
import cz.fidentis.analyst.project.Project;
import cz.fidentis.analyst.rendering.Camera;

import javax.swing.*;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.ObjectInputStream;
import java.util.function.BiConsumer;

/**
 * A worker for loading multiple task tabs (faces and cameras) of a project from dumped files.
 * 
 * @author Matej Kovar
 */
public class RecoverTasksWorker extends SwingWorker<Boolean, Integer> {
    
    private Project project;
    private ProgressDialog<Boolean, Integer> progressDialog;
    private final BiConsumer<Camera, FacesProxy> consumer;
    
    /**
     * Constructor.
     * 
     * @param project Project with info about project folder containing stored tasks
     * @param progressDialog A progress dialog window
     * @param consumer A method that creates an analytical tab from loaded camera and faces
     * @throws IllegalArgumentException if the project has not a folder assigned
     */
    public RecoverTasksWorker(Project project, ProgressDialog<Boolean, Integer> progressDialog, BiConsumer<Camera, FacesProxy> consumer) {
        if (!project.hasProjectDir()) {
            throw new IllegalArgumentException("project");
        }
        this.project = project;
        this.progressDialog = progressDialog;
        this.consumer = consumer;
    }

    @Override
    protected Boolean doInBackground() throws Exception {
        File projectDir = new File(project.getProjectDir().toString());

        File[] files = projectDir.listFiles((dir, name) -> name.toLowerCase().endsWith(".bin"));

        int counter = 0;
        for (File file: files) {
            if (isCancelled()) {
                return false;
            }

            try (ObjectInputStream in = new ObjectInputStream(new BufferedInputStream(new FileInputStream(file)))) {
                Camera camera = (Camera) in.readObject();
                FacesProxy faces = new FacesProxy(in);
                this.consumer.accept(camera, faces);
            } catch (Exception ex) {
                Logger.print("[ERROR] Task recovery: " + ex);
                throw ex;
            }

            int progress = (int)Math.round(100.0 * ++counter / files.length);
            progressDialog.setValue(progress);
        }

        return true;
    }
    
    @Override
    protected void done() {
        progressDialog.dispose();
    }
}
