package cz.fidentis.analyst.gui.project;

import cz.fidentis.analyst.gui.elements.ProgressDialog;
import cz.fidentis.analyst.gui.task.TaskWindow;
import cz.fidentis.analyst.project.Project;

import javax.swing.*;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.util.List;

/**
 * A worker for saving  the state of task tabs into the disk.
 * 
 * @author Radek Oslejsek
 */
public class SaveTasksWorker extends SwingWorker<Boolean, Integer> {
    
    private final Project project;
    private final List<TaskWindow> taskTabs;
    private final ProgressDialog progressDialog;
    
    /**
     * Constructor.
     * 
     * @param project Project with info about project folder containing stored tasks
     * @param taskTabs A list of analytical tasks to be saved
     * @param progressDialog Dialog window
     * @throws IllegalArgumentException if the project has not a folder assigned
     */
    public SaveTasksWorker(Project project, List<TaskWindow> taskTabs, ProgressDialog progressDialog) {
        if (!project.hasProjectDir()) {
            throw new IllegalArgumentException("project");
        }
        this.taskTabs = taskTabs;
        this.project = project;
        this.progressDialog = progressDialog;
    }
    
    /**
     * Returns path to the task's file.
     * 
     * @param project Project
     * @param taskName Name of the task
     * @return path to the task's file.
     */
    public static File getTaskFile(Project project, String taskName) {
        String fileName = taskName.replaceAll("\\s", "-").replaceAll(":", "");
        String projectDir = project.getProjectDir().toString() + File.separatorChar;
        return new File(projectDir + "task-" + fileName + ".bin");
    }

    @Override
    protected Boolean doInBackground() throws Exception {
        int counter = 0;
        for (TaskWindow task : taskTabs) {
            if (isCancelled()) {
                return false;
            }
            
            File file = getTaskFile(this.project, task.getName());
            try (ObjectOutputStream out = new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream(file)))) {
                out.writeObject(task.getCanvas().getCamera());
                task.getFacesProxy().serialize(out);
            }
            
            int progress = (int)Math.round(100.0 * ++counter / taskTabs.size());
            progressDialog.setValue(progress);
        }
        
        return true;
    }
    
    @Override
    protected void done() {
        progressDialog.dispose();
    }
}
