package cz.fidentis.analyst.gui.task;

import cz.fidentis.analyst.canvas.Canvas;
import cz.fidentis.analyst.gui.task.batch.FacesProxyDecorator;
import cz.fidentis.analyst.gui.task.batch.distance.BatchDistanceAction;
import cz.fidentis.analyst.gui.task.batch.registration.BatchRegistrationAction;
import cz.fidentis.analyst.gui.task.curvature.CurvatureAction;
import cz.fidentis.analyst.gui.task.distance.DistanceAction;
import cz.fidentis.analyst.gui.task.faceinfo.FaceOverviewAction;
import cz.fidentis.analyst.gui.task.featurepoints.FeaturePointsAction;
import cz.fidentis.analyst.gui.task.interactivemask.InteractiveMaskAction;
import cz.fidentis.analyst.gui.task.registration.RegistrationAction;
import cz.fidentis.analyst.gui.task.symmetry.CuttingPlanesAction;
import cz.fidentis.analyst.gui.task.symmetry.SymmetryAction;
import cz.fidentis.analyst.project.FacesProxy;
import cz.fidentis.analyst.rendering.Camera;
import cz.fidentis.analyst.rendering.Scene;
import cz.fidentis.analyst.toolbar.SceneToolboxFaceToFace;
import cz.fidentis.analyst.toolbar.SceneToolboxSingleFace;
import org.openide.windows.TopComponent;

import javax.swing.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.util.Objects;

/**
 * The non-singleton window/tab for the analysis of one, two or many to many faces
 *
 * @author Matej Kovar
 */
@TopComponent.Description(
        preferredID = "TaskWindow",
        iconBase="analysis.png",
        persistenceType = TopComponent.PERSISTENCE_NEVER
)
public class TaskWindow extends TopComponent {
    
    private final Canvas canvas ;
    private final TaskControlPane controlPanel;
    private final JScrollPane scrollPane;
    
    private final transient FacesProxy faces;
    
    private final transient ActionListener endTaskListener;
    
    private boolean askWhenClosing = true;
    
    /**
     * Constructor.
     * 
     * @param faces faces of analytical task
     * @param endTaskListener action endTaskListener
     */
    public TaskWindow(FacesProxy faces, ActionListener endTaskListener) {
        this.faces = faces;
        this.canvas = new Canvas();
        this.endTaskListener = endTaskListener;

        initCanvas();

        controlPanel = new TaskControlPane();
        scrollPane = new JScrollPane(controlPanel);

        setName(faces.getTaskName());
        initComponents();

        // change the height so that it corresponds to the height of the OpenGL window
        canvas.addComponentListener(new ComponentAdapter() {
            @Override
            public void componentResized(ComponentEvent e) {
                scrollPane.setSize(ControlPanel.CONTROL_PANEL_WIDTH, canvas.getHeight());
            }
        });

        controlPanel.addChangeListener(e -> getCanvas().renderScene());

        initTabs();
    }
    
    /**
     * Determines whether a dialogue window is opened when the window is closing.
     * @param ask If {@code true}, then dialogue window pups up
     */
    public void setAskWhenClosing(boolean ask) {
        this.askWhenClosing = ask;
    }
    
    @Override
    public boolean canClose() {
        if (!askWhenClosing) {
            return true;
        }
        int answer = JOptionPane.showConfirmDialog(
                this,
                "Do you really want to end this analytical task?\nChanges will be discarded!", 
                "", 
                JOptionPane.YES_NO_OPTION, 
                JOptionPane.WARNING_MESSAGE
        );
        
        if (answer == JOptionPane.YES_OPTION && super.canClose()) {
            // notify project panel about closing me
            endTaskListener.actionPerformed(new ActionEvent(
                    this,
                    ActionEvent.ACTION_PERFORMED,
                    ""
            ));
            return true;
        } else {
            return false;
        }
    }
    
    @Override
    public int getPersistenceType() {
        return TopComponent.PERSISTENCE_NEVER; 
    }
    
    /**
     * Checks whether this tab contains a face with given name
     * @param name String name of face
     * @return true if face with this name is in this tab
     */
    public boolean hasFace(String name) {
        for (int i = 0; i < faces.getNumFaces(); i++) {
            if (name.equals(faces.getFace(i).getShortName())) {
                return true;
            }
        }
        return false;
    }

    public Canvas getCanvas() {
        return canvas;
    }
    
    public FacesProxy getFacesProxy() {
        return this.faces;
    }
    
    /**
     * Sets camera to canvas
     * @param camera Camera
     */
    public void setCamera(Camera camera) {
        this.canvas.setCamera(camera);
    }

    @Override
    public int hashCode() {
        return getName().hashCode();
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final TaskWindow other = (TaskWindow) obj;
        
        return Objects.equals(this.getName(), other.getName());
    }

    private void initCanvas() {
        if (faces.getNumFaces() <= 2) { // initialize primary face
            Scene scene = canvas.getScene();
            int index = scene.getFreeSlotForFace();
            scene.setHumanFace(index, faces.getPrimaryFace());
            scene.setFaceAsPrimary(index);
            faces.getPrimaryFace().registerListener(canvas);
        }
        
        if (faces.getNumFaces() == 2) { // initialize secondary face
            Scene scene = canvas.getScene();
            int index = scene.getFreeSlotForFace();
            scene.setHumanFace(index, faces.getSecondaryFace());
            scene.setFaceAsSecondary(index);
            faces.getSecondaryFace().registerListener(canvas);
        }
        
        canvas.getScene().setDefaultColors();
        
        if (faces.getNumFaces() == 2) {
            canvas.addToolBox(new SceneToolboxFaceToFace(canvas, faces));
        } else { // one or more than two
            canvas.getScene().setDefaultColors();
            canvas.addToolBox(new SceneToolboxSingleFace(canvas, faces));
        }

        canvas.getCamera().zoomToFit(canvas.getScene());
    }
    
    private void initComponents() {
        GroupLayout layout = new GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
                layout.createParallelGroup(GroupLayout.Alignment.LEADING)
                        .addGroup(layout.createSequentialGroup()
                                .addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
                                .addComponent(canvas, GroupLayout.DEFAULT_SIZE, 651, Short.MAX_VALUE)
                                .addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
//                                .addComponent(renderingToolBar, GroupLayout.PREFERRED_SIZE, RenderingToolBar.WIDTH, GroupLayout.PREFERRED_SIZE)
                                .addComponent(
                                        scrollPane,
                                        ControlPanel.CONTROL_PANEL_WIDTH, 
                                        ControlPanel.CONTROL_PANEL_WIDTH, 
                                        ControlPanel.CONTROL_PANEL_WIDTH
                                )
                        )
        );
        layout.setVerticalGroup(
                layout.createParallelGroup(GroupLayout.Alignment.LEADING)
                        .addGroup(layout.createSequentialGroup()
                                .addGroup(layout.createBaselineGroup(true, true)
                                        .addComponent(canvas)
//                                        .addComponent(renderingToolBar)
                                        .addComponent(scrollPane)
                                ))
        );
    }

    private void initTabs() {
        switch (faces.getNumFaces()) {
            case 1:
                new SymmetryAction(getCanvas(), faces, controlPanel).popup();
                new CuttingPlanesAction(getCanvas(), faces, controlPanel);
                new FeaturePointsAction(getCanvas(), faces, controlPanel);
                new CurvatureAction(getCanvas(), faces, controlPanel);
                new FaceOverviewAction(getCanvas(), faces, controlPanel, 0);
                new InteractiveMaskAction(getCanvas(), faces, controlPanel);
                break;
            case 2:
                new RegistrationAction(canvas, faces, controlPanel).popup();
                new DistanceAction(canvas, faces, controlPanel);
                new SymmetryAction(canvas, faces, controlPanel);
                new CuttingPlanesAction(canvas, faces, controlPanel);
                new FaceOverviewAction(getCanvas(), faces, controlPanel, 0);
                new FaceOverviewAction(getCanvas(), faces, controlPanel, 1);
                //new InteractiveMaskAction(getCanvas(), faces, controlPanel);
                break;
            default: // batch mode
                FacesProxyDecorator batchFacesProxy = new FacesProxyDecorator(faces);
                new BatchRegistrationAction(canvas, batchFacesProxy, controlPanel);
                new BatchDistanceAction(canvas, batchFacesProxy, controlPanel);
                //new BatchCuttingPlanesAction(canvas, batchFacesProxy, controlPanel);
                break;
        }
    }
    
}