package cz.fidentis.analyst.gui.task.batch;

import cz.fidentis.analyst.data.face.HumanFace;
import cz.fidentis.analyst.data.face.HumanFaceMemoryManager;
import cz.fidentis.analyst.project.FacesProxy;

import javax.swing.*;
import java.io.File;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.List;
import java.util.Objects;

/**
 * In contrast to the {@code FacesProxy}, which deals with faces that are the subject of batch processing,
 * this class add handling of the (optionally computed) average face.
 * While the faces can be temporarily dumped, the average face is always in the memory.
 * It is assumed that the average face is added as the first item of a menu (combo box), whenever is computed.
 * This class provides methods for such menu (combo box) management and synchronization.
 *
 * @author Radek Oslejsek
 */
public class FacesProxyDecorator extends FacesProxy {
    private HumanFace avgFace;

    private final FacesProxy facesProxy;

    /**
     * Constructor.
     *
     * @param facesProxy original faces proxy dealing with faces to be processed
     */
    public FacesProxyDecorator(FacesProxy facesProxy) {
        Objects.requireNonNull(facesProxy);
        this.facesProxy = facesProxy;
    }

    public FacesProxy getFacesProxy() {
        return facesProxy;
    }

    /**
     * Returns the average human face or {@code null}
     * @return the average human face or {@code null}
     */
    public HumanFace getAvgFace() {
        return avgFace;
    }

    /**
     * Fill the menu (combo box) with names of faces, including the average face, if provided as input parameter.
     * The average face is stored.
     * An action event is triggered automatically by the {@code cbox} with with the originally selected item or
     * the newly added average face.
     *
     * @param avgFace The average face, can be {@code null}
     * @param selectNewAvgFace If {@code true} and a new average is added, then the average face is automatically selected.
     *                         If the {@code avgFace} is {@code null}, then this parameter has no effect.
     * @param cbox A combo box (menu)
     */
    public void syncSelectionMenu(HumanFace avgFace, boolean selectNewAvgFace, JComboBox<String> cbox) {
        this.avgFace = avgFace;
        syncSelectionMenu(selectNewAvgFace, cbox);
    }

    /**
     * Fill the menu (combo box) with names of faces stored in the face proxy, plus the optional average face.
     * An action event is triggered automatically by the {@code cbox} with the originally selected item or
     * the newly added average face.
     *
     * @param selectNewAvgFace If {@code true} and a new average is added, then the average face is automatically selected.
     * @param cbox A combo box (menu)
     */
    public void syncSelectionMenu(boolean selectNewAvgFace, JComboBox<String> cbox) {
        int expectedNumItems = haveAvgFace() ? facesProxy.getNumFaces() + 1 : facesProxy.getNumFaces();
        int realNumItems = cbox.getItemCount();
        int selectedItem = cbox.getSelectedIndex();

        cbox.removeAllItems();
        if (haveAvgFace()) {
            cbox.addItem("Average face");
        }
        facesProxy.getFaceFiles().forEach(f -> {
            String name = f.toString();
            name = name.substring(name.lastIndexOf(File.separatorChar) + 1);
            cbox.addItem(name);
        });

        // re-select proper item:
        if (realNumItems < expectedNumItems) { // adding a new average face
            if (selectNewAvgFace) {
                cbox.setSelectedIndex(0);
            } else if (selectedItem >= 0) {
                cbox.setSelectedIndex(selectedItem+1);
            }
        } else if (realNumItems > expectedNumItems) { // removing the average face
            cbox.setSelectedIndex((--selectedItem >=0) ? selectedItem : 0);
        } else { // nothing has changed
            if (selectedItem >= 0) {
                cbox.setSelectedIndex(selectedItem);
            }
        }
    }

    /**
     * Returns the face selected in a menu.
     *
     * @param cbox A combo box (menu)
     * @return The face selected in a menu. This is either some face from the dataset, or the average face.
     */
    public HumanFace getSelectedFace(JComboBox<String> cbox) {
        if (cbox.getItemCount() == 0) {
            return null;
        }
        if (isAvgFaceSelected(cbox)) {
            return avgFace;
        }
        int index = haveAvgFace() ? cbox.getSelectedIndex()-1 : cbox.getSelectedIndex();
        String id = facesProxy.getFactory().loadFace(facesProxy.getFaceFiles().get(index));
        return facesProxy.getFactory().getFace(id);
    }

    /**
     * Determines whether the average face has been already computed.
     *
     * @return {@code true} if the average face has been already computed
     */
    public boolean haveAvgFace() {
        return avgFace != null;
    }
    
    /**
     * Returns {@code true} if the item selected in the combo box is the average face
     *
     * @param cbox A combo box (menu)
     * @return {@code true} if the item selected in the combo box is the average face
     */
    public boolean isAvgFaceSelected(JComboBox<String> cbox) {
        return haveAvgFace() && cbox.getSelectedIndex() == 0;
    }


    // ADAPTED PUBLIC METHODS of the FacesProxy class:

    @Override
    public void serialize(ObjectOutputStream out) throws IOException {
        facesProxy.serialize(out);
    }

    @Override
    public List<File> getFaceFiles() {
        return facesProxy.getFaceFiles();
    }

    @Override
    public File getFaceFile(int index) {
        return facesProxy.getFaceFile(index);
    }

    @Override
    public File getFaceFile(String faceName) {
        return facesProxy.getFaceFile(faceName);
    }

    @Override
    public HumanFace getFace(int index) {
        return facesProxy.getFace(index);
    }

    @Override
    public int getFaceIndex(String faceName) {
        return facesProxy.getFaceIndex(faceName);
    }

    @Override
    public int addFace(File file, boolean unique) {
        return facesProxy.addFace(file, unique);
    }

    @Override
    public boolean removeFace(String faceName) {
        return facesProxy.removeFace(faceName);
    }

    @Override
    public boolean isInMemory(int index) {
        return facesProxy.isInMemory(index);
    }

    @Override
    public String getFaceName(int index) {
        return facesProxy.getFaceName(index);
    }

    @Override
    public String getFacePath(int index, FaceFileType type) {
        return facesProxy.getFacePath(index,type);
    }

    @Override
    public String getFaceDir(int index) {
        return facesProxy.getFaceDir(index);
    }

    @Override
    public HumanFace getPrimaryFace() {
        return facesProxy.getPrimaryFace();
    }

    @Override
    public HumanFace getSecondaryFace() {
        return facesProxy.getSecondaryFace();
    }

    @Override
    public int getNumFaces() {
        return facesProxy.getNumFaces();
    }

    @Override
    public String getTaskName() {
        return facesProxy.getTaskName();
    }

    @Override
    public boolean equals(Object obj) {
        return facesProxy.equals(obj);
    }

    @Override
    public int hashCode() {
        return facesProxy.hashCode();
    }

    @Override
    public HumanFaceMemoryManager getFactory() {
        return facesProxy.getFactory();
    }
}
