package cz.fidentis.analyst.gui.task.batch.distance;

import cz.fidentis.analyst.data.face.HumanFace;
import cz.fidentis.analyst.gui.elements.ProgressDialog;
import javax.swing.SwingWorker;

/**
 * A task that computes distance-based similarity of the set of faces.
 * 
 * @author Radek Oslejsek
 */
public abstract class BatchDistanceTask extends SwingWorker<Void, Integer> {
    
    private final ProgressDialog<Void, Integer> progressDialog;
    private final BatchDistancePanel controlPanel;

    private double[][] distances; // AVG distance
    
    private double[][] deviations; // sample standard deviation
    
    
    /**
     * Constructor.
     * 
     * @param progressDialog A window that show the progress of the computation. Must not be {@code null}
     * @param controlPanel A control panel with computation parameters. Must not be {@code null}
     * The average face is skipped from the similarity measurement (i.e., we do not measure its distance to other distances)
     */
    protected BatchDistanceTask(ProgressDialog<Void, Integer> progressDialog, BatchDistancePanel controlPanel) {
        this.progressDialog = progressDialog;
        this.controlPanel = controlPanel;
        int nFaces = getControlPanel().getFacesProxyDecorator().getFaceFiles().size();
        distances = new double[nFaces][nFaces];
        deviations = new double[nFaces][nFaces];
    }
    
    @Override
    protected void done() {
        progressDialog.dispose(); // close progress bar
        if (isCancelled()) {
            distances = null;
            deviations = null;
        }
    }
    
    /**
     * Returns computed 2D matrix of distance similarities or {@code null}
     * @return computed 2D matrix of distance similarities or {@code null}
     */
    public double[][] getDistSimilarities() {
        return distances;
    }
    
    /**
     * Returns computed 2D matrix of sample standard deviations or {@code null}
     * @return computed 2D matrix of sample standard deviations or {@code null}
     */
    public double[][] getDistDeviations() {
        return this.deviations;
    }
    
    public ProgressDialog<Void, Integer> getProgressDialog() {
        return progressDialog;
    }

    public final BatchDistancePanel getControlPanel() {
        return controlPanel;
    }

    protected HumanFace getFaceFromFaceFactory(int faceIndex) {
        String faceId = controlPanel
                .getFacesProxyDecorator()
                .getFactory()
                .loadFace(controlPanel.getFacesProxyDecorator().getFaceFiles().get(faceIndex));
        return controlPanel.getFacesProxyDecorator().getFactory().getFace(faceId);
    }

    protected HumanFace getSelectedFace() {
        return getControlPanel().getSelectedFace();
    }
    
    protected void setDistSimilarity(int i, int j, double val) {
        this.distances[i][j] = val;
    }
    
    protected void setDistDeviation(int i, int j, double val) {
        this.deviations[i][j] = val;
    }
}
