package cz.fidentis.analyst.gui.task.batch.distance;

import cz.fidentis.analyst.Logger;
import cz.fidentis.analyst.data.face.HumanFace;
import cz.fidentis.analyst.data.face.HumanFaceMemoryManager;
import cz.fidentis.analyst.engines.distance.MeshDistanceConfig;
import cz.fidentis.analyst.engines.distance.MeshDistanceVisitor;
import cz.fidentis.analyst.engines.distance.measurement.DistanceRecord;
import cz.fidentis.analyst.engines.face.FaceStateServices;
import cz.fidentis.analyst.gui.elements.ProgressDialog;
import cz.fidentis.analyst.gui.task.batch.Stopwatch;
import org.apache.commons.lang3.tuple.ImmutablePair;

import javax.vecmath.Point3d;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.*;
import java.util.logging.Level;

/**
 * A task that computes similarity of a set of faces by computing
 * the distance of faces to an average face and then combining these values
 * to get mutual similarity for all pairs.
 * <p>
 * The computation is accelerated by using multiple CPU cores concurrently.
 * The exact computation parameters are taken from the {@code BatchDistancePanel}.
 * </p>
  * 
 * @author Radek Oslejsek
 */
public class NearestNeighborsDistanceTask extends BatchDistanceTask {
    
    /**
     * The strategy of distance measurement.
     * @author Radek Oslejsek
     */
    public enum DistMeasurement {
        RELATIVE_DISTANCE,
        VECTOR_DISTANCE,
        COMBINED
    }
    
    private final Stopwatch totalTime = new Stopwatch("Total computation time:\t\t");
    private final Stopwatch distComputationTime = new Stopwatch("Distance computation time:\t");
    private final Stopwatch loadTime = new Stopwatch("Disk access time:\t\t");
    private final Stopwatch kdTreeConstructionTime = new Stopwatch("KD trees construction time:\t\t");
    private final Stopwatch finalDistComputationTime = new Stopwatch("Distance finalization time:\t");
    
    private final DistMeasurement distMeasurement;
    
    /**
     * Constructor.
     * 
     * @param progressDialog A window that show the progress of the computation. Must not be {@code null}
     * @param controlPanel A control panel with computation parameters. Must not be {@code null}
     * @param distMeasurement A strategy of measuring indirect distance
     */
    public NearestNeighborsDistanceTask(ProgressDialog<Void, Integer> progressDialog, BatchDistancePanel controlPanel, DistMeasurement distMeasurement) {
        super(progressDialog, controlPanel);
        this.distMeasurement = distMeasurement;
    }
    
    @Override
    protected Void doInBackground() throws Exception {
        HumanFaceMemoryManager factory = getControlPanel().getFacesProxyDecorator().getFactory();
        factory.setReuseDumpFile(true); // it's safe because no changes are made to models 
        factory.setStrategy(HumanFaceMemoryManager.Strategy.MRU); // keep first X faces in the memory

        totalTime.start();

        // We don't need to reload the selected "gauge" face periodically for two reasons:
        //   - It is never dumped from memory to disk because we use MRU
        //   - Even if dumped, the face keeps in the memory until we hold the pointer to it
        loadTime.start();
        HumanFace gaugeFace = getSelectedFace();
        loadTime.stop();
        
        // Cache of distances of individual vertices
        //   distCache.get(i) = i-th face
        //   distCache.get(i).get(j) = distance of j-th vertex of i-th face
        DistCache distCache = null;
        switch (distMeasurement) {
            case RELATIVE_DISTANCE:
                distCache = new RelativeDistCache();
                break;
            case VECTOR_DISTANCE:
                distCache = new DirectionCache();
                break;
            default:
            case COMBINED:
                distCache = new CombinedCache();
                break;
        }

        int numFaces = getControlPanel().getFacesProxyDecorator().getFaceFiles().size();
        for (int i = 0; i < numFaces; i++) {

            if (isCancelled()) { // the user canceled the process
                return null;
            }
            
            loadTime.start();
            HumanFace face = getFaceFromFaceFactory(i);
            loadTime.stop();
                
            kdTreeConstructionTime.start();
            FaceStateServices.updateKdTree(face, FaceStateServices.Mode.COMPUTE_IF_ABSENT);
            kdTreeConstructionTime.stop();

            distComputationTime.start();
            MeshDistanceVisitor dist = new MeshDistanceConfig(
                    MeshDistanceConfig.Method.POINT_TO_POINT_NEAREST_NEIGHBORS,
                    face.getKdTree(),
                    true,
                    true).getVisitor();
            gaugeFace.getMeshModel().compute(dist);
            distCache.addValuesOfFace(dist);
            distComputationTime.stop();
                
            // Memory is more valuable. And it's probable that we will not need
            // k-d trees anymore. If yes, then they are quickly re-computed.
            FaceStateServices.updateKdTree(face, FaceStateServices.Mode.DELETE);

            // update progress bar
            getProgressDialog().setValue((int) Math.round(100.0 * (i+1) / numFaces));
        }
        
        finalDistComputationTime.start();
        finalizeDistance(distCache);
        finalDistComputationTime.stop();
        
        totalTime.stop();

        printTimeStats();

        return null;
    }
    
    protected void printTimeStats() {
        Logger.print(distComputationTime.toString());
        Logger.print(finalDistComputationTime.toString());
        Logger.print(loadTime.toString());
        Logger.print(kdTreeConstructionTime.toString());
        Logger.print(totalTime.toString());
    }
    
    private void finalizeDistance(DistCache distCache) {
        ExecutorService executor = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
        final List<Future<Void>> results = new ArrayList<>();
        
        for (int i = 0; i < distCache.getNumFaces(); i++) {
            for (int j = i; j < distCache.getNumFaces(); j++) {
                //assert(distCache.get(i).size() == distCache.get(i).size()); // should never happen
                
                final int fi = i;
                final int fj = j;
                
                // Compute average HD for the pair of faces in a separate threat:
                results.add(executor.submit(new Callable<Void>() {
                    @Override
                    public Void call() throws Exception {
                        if (fi == fj) {
                            setDistSimilarity(fi, fj, 0.0);
                            setDistSimilarity(fj, fi, 0.0);
                            setDistDeviation(fi, fj, 0.0);
                            setDistDeviation(fj, fi, 0.0);
                            return null;
                        }
                        
                        //double max = Double.NEGATIVE_INFINITY;
                        double sum = 0.0;
                        int counter = 0;
                        for (int k = 0; k < distCache.getNunFaceVertices(fi); k++) {
                            double d = distCache.getDistance(fi, fj, k);
                            if (Double.isFinite(d)) {
                                sum += d;
                                counter++;
                            }
                        }
                        double avg = sum / counter;
                        double dev = Math.sqrt((1.0 / (counter-1)) * sum);
                        setDistSimilarity(fi, fj, avg);
                        setDistSimilarity(fj, fi, avg);
                        //setDistSimilarity(fi, fj, max);
                        //setDistSimilarity(fj, fi, max);
                        setDistDeviation(fi, fj, dev);
                        setDistDeviation(fj, fi, dev);
                        return null;
                    }
                }));
            }
        }
        
        executor.shutdown();
        while (!executor.isTerminated()){}
        try {
            for (Future<Void> res: results) {
                res.get(); // waits until all computations are finished
            }
        } catch (final InterruptedException | ExecutionException ex) {
            java.util.logging.Logger.getLogger(NearestNeighborsDistanceTask.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
    
    
    /**
     * Cache of distances from the average face to other faces.
     * 
     * @param <T> the type of cached value
     * @author Radek Oslejsek
     */
    private abstract class DistCache<T> {
         private final List<List<T>> cache = new ArrayList<>();
         
         public abstract void addValuesOfFace(MeshDistanceVisitor distVisitor);
         
         public abstract double getDistance(int faceIndex1, int faceIndex2, int vertIndex);

         /**
          * Returns the number of cached faces
          * @return the number of cached faces
          */
         public int getNumFaces() {
             return cache.size();
         }
         
         /**
          * Returns the number of cached vertices of the given faces
          * @return the number of cached vertices of the given faces
          */
         public int getNunFaceVertices(int faceIndex) {
            return cache.get(faceIndex).size();
         }
         
         protected List<List<T>> getCache() {
             return Collections.unmodifiableList(cache);
         }
         
         protected void add(List<T> list) {
             cache.add(list);
         }
    }
    
    /**
     * Cache of relative distances from the average face to other faces.
     * 
     * @author Radek Oslejsek
     */
    private class RelativeDistCache extends DistCache<Double> {
        
        @Override
        public void addValuesOfFace(MeshDistanceVisitor distVisitor) {
            if (distVisitor == null) {
                add(null);
            } else { // Store relative distances of vertices to the cache
                add(distVisitor.getDistancesOfVisitedFacets().stream()
                        .mapToDouble(DistanceRecord::getDistance)
                        .boxed()
                        .toList());
            }
        }
        
        @Override
        public double getDistance(int faceIndex1, int faceIndex2, int vertIndex) {
            double d1 = getCache().get(faceIndex1).get(vertIndex);
            double d2 = getCache().get(faceIndex2).get(vertIndex);
            return (Double.isFinite(d1) && Double.isFinite(d2))
                    ? Math.abs(d1 - d2)
                    : Double.NaN;
        }
    }
    
    /**
     * Cache of vectors to the nearest neighbors, from which the final distance is computed.
     * 
     * @author Radek Oslejsek
     */
    private class DirectionCache extends DistCache<Point3d> {
        
        @Override
        public void addValuesOfFace(MeshDistanceVisitor distVisitor) {
            if (distVisitor == null) {
                add(null);
            } else {
                add(distVisitor.getDistancesOfVisitedFacets().stream()
                        .map(DistanceRecord::getNearestNeighbor)
                        .map(mp -> mp == null ? null : mp.getPosition())
                        .toList());
            }
        }
        
        @Override
        public double getDistance(int faceIndex1, int faceIndex2, int vertIndex) {
            Point3d p1 = getCache().get(faceIndex1).get(vertIndex);
            Point3d p2 = getCache().get(faceIndex2).get(vertIndex);
            return (p1 != null && p2 != null)
                    ? p1.distance(p2)
                    : Double.NaN;
        }
    }
    
    /**
     * TO DO
     * 
     * @author Radek Oslejsek
     */
    private class CombinedCache extends DistCache<ImmutablePair<Point3d, Double>> {
        
        @Override
        public void addValuesOfFace(MeshDistanceVisitor distVisitor) {
            if (distVisitor == null) {
                add(null);
            } else {
                List<Point3d> nearestPoints = distVisitor.getDistancesOfVisitedFacets().stream()
                        .map(DistanceRecord::getNearestNeighbor)
                        .map(mp -> mp == null ? null : mp.getPosition())
                        .toList();
                
                List<Double> distances = distVisitor.getDistancesOfVisitedFacets().stream()
                        .map(DistanceRecord::getDistance)
                        .toList();
                
                assert(nearestPoints.size() == distances.size());
                
                List<ImmutablePair<Point3d, Double>> list = new ArrayList<>(distances.size());
                for (int i = 0; i < distances.size(); i++) {
                    list.add(new ImmutablePair<>(nearestPoints.get(i), distances.get(i)));
                }
                add(list);
            }
        }
        
        @Override
        public double getDistance(int faceIndex1, int faceIndex2, int vertIndex) {
            double d1 = getCache().get(faceIndex1).get(vertIndex).getRight();
            double d2 = getCache().get(faceIndex2).get(vertIndex).getRight();
            
            if (!Double.isFinite(d1) || !Double.isFinite(d2)) {
                return Double.NaN;
            } else if (d1 >= 0 && d2 >= 0 || d1 <= 0 && d2 <= 0) {
                return Math.abs(d1 - d2);
            } else {            
                Point3d p1 = getCache().get(faceIndex1).get(vertIndex).getLeft();
                Point3d p2 = getCache().get(faceIndex2).get(vertIndex).getLeft();
                return (p1 != null && p2 != null)
                        ? p1.distance(p2)
                        : Double.NaN;
            }
        }
    }
}
