package cz.fidentis.analyst.gui.task.batch.distance;

import cz.fidentis.analyst.Logger;
import cz.fidentis.analyst.data.face.HumanFace;
import cz.fidentis.analyst.data.face.HumanFaceMemoryManager;
import cz.fidentis.analyst.data.kdtree.KdTree;
import cz.fidentis.analyst.data.mesh.MeshFacet;
import cz.fidentis.analyst.engines.distance.MeshDistanceConfig;
import cz.fidentis.analyst.engines.distance.MeshDistanceVisitor;
import cz.fidentis.analyst.engines.face.FaceStateServices;
import cz.fidentis.analyst.gui.elements.ProgressDialog;
import cz.fidentis.analyst.gui.task.batch.Stopwatch;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * A task that computes similarity of a set of faces by applying two-way (from A to B and vice versa)
 * distance measurement to all pairs A and B if the set.
 * The average face, if exists, is skipped from the similarity measurement
 * (i.e., we do not measure its distance to other faces).
 * The exact computation parameters are taken from the {@code BatchDistancePanel}.
 * 
 * @author Radek Oslejsek
 */
public class PairwiseDistanceTask extends BatchDistanceTask {

    private final Stopwatch totalTime = new Stopwatch("Total computation time:\t");
    private final Stopwatch distComputationTime = new Stopwatch("Distance measurement time:\t");
    private final Stopwatch loadTime = new Stopwatch("Disk access time:\t");
    
    private final boolean crop;
    private KdTree meanFaceKdTree;
    
    /**
     * Constructor.
     * 
     * @param progressDialog A window that show the progress of the computation. Must not be {@code null}
     * @param controlPanel A control panel with computation parameters. Must not be {@code null}
     * @param crop If {@code true}, then the measured faces are auto-cropped to the surface of the average face.
     *        This option slows down the computation even more, but enables us to compare 
     *        the results of this pairwise comparison other methods based of indirect measurement
     *        (i.e., the methods that use the average face as a "gauge").
     *        This option is ignored if there is average face computed.
     **/
    public PairwiseDistanceTask(ProgressDialog<Void, Integer> progressDialog, BatchDistancePanel controlPanel, boolean crop) {
        super(progressDialog, controlPanel);
        this.crop = crop && controlPanel.getFacesProxyDecorator().haveAvgFace();
    }

    @Override
    protected Void doInBackground() throws Exception {
        HumanFaceMemoryManager factory = getControlPanel().getFacesProxyDecorator().getFactory();
        factory.setReuseDumpFile(true); // it's safe because no changes are made to models
        factory.setStrategy(HumanFaceMemoryManager.Strategy.MRU);
        
        totalTime.start();
        
        if (this.crop) {
            FaceStateServices.updateKdTree(
                    getControlPanel().getFacesProxyDecorator().getAvgFace(),
                    FaceStateServices.Mode.COMPUTE_ALWAYS);
            this.meanFaceKdTree = getControlPanel().getFacesProxyDecorator().getAvgFace().getKdTree();
        }
        
        int counter = 0;
        int numFaces = getControlPanel().getFacesProxyDecorator().getFaceFiles().size();
        for (int i = 0; i < numFaces; i++) {
            for (int j = i; j < numFaces; j++) { // starts with "i"!
                
                if (isCancelled()) { // the user canceled the process
                    return null;
                }

                // (Re-)load the primary face (it could be dumped in the meantime) and load the second face
                loadTime.start();
                HumanFace priFace = getFaceFromFaceFactory(i);
                HumanFace secFace = getFaceFromFaceFactory(j);
                loadTime.stop();
                
                // compute distance of current face to the mean face:
                MeshDistanceVisitor meanDistVisitor = null;
                if (crop) {
                    meanDistVisitor = new MeshDistanceConfig(
                            MeshDistanceConfig.Method.POINT_TO_POINT_NEAREST_NEIGHBORS,
                            this.meanFaceKdTree,
                            false,
                            true).getVisitor();
                    secFace.getMeshModel().compute(meanDistVisitor);
                }
                
                // compute Hausdorff distance in both directions
                distComputationTime.start();
                FaceStateServices.updateKdTree(priFace, FaceStateServices.Mode.COMPUTE_IF_ABSENT);
                FaceStateServices.updateKdTree(secFace, FaceStateServices.Mode.COMPUTE_IF_ABSENT);

                MeshDistanceVisitor dist = new MeshDistanceConfig(
                        MeshDistanceConfig.Method.POINT_TO_POINT_NEAREST_NEIGHBORS,
                        priFace.getKdTree(),
                        false,
                        true).getVisitor();
                secFace.getMeshModel().compute(dist);
                setDistValues(j, i, dist, meanDistVisitor);
                
                // compute distance of current face to the mean face:
                meanDistVisitor = null;
                if (crop) {
                     meanDistVisitor = new MeshDistanceConfig(
                             MeshDistanceConfig.Method.POINT_TO_POINT_NEAREST_NEIGHBORS,
                             this.meanFaceKdTree,
                             false,
                             true).getVisitor();
                    priFace.getMeshModel().compute(meanDistVisitor);
                }
                dist = new MeshDistanceConfig(
                        MeshDistanceConfig.Method.POINT_TO_POINT_NEAREST_NEIGHBORS,
                        secFace.getKdTree(),
                        false,
                        true).getVisitor();
                priFace.getMeshModel().compute(dist);
                setDistValues(i, j, dist, meanDistVisitor);
                distComputationTime.stop();
                
                // update progress bar
                getProgressDialog().setValue((int) Math.round(100.0 * ++counter / numCycles(numFaces)));
            }
        }
        
        totalTime.stop();

        printTimeStats();

        return null;
    }
    
    protected void printTimeStats() {
        Logger.print(distComputationTime.toString());
        Logger.print(loadTime.toString());
        Logger.print(totalTime.toString());
    }
    
    protected double numCycles(int nFaces) {
        return 0.5 * nFaces * (nFaces + 1);
    }
    
    /**
     * Sets an average distance and a sample standard deviation but only considering
     * vertices that are not copped either by the second face in the pairwise comparison
     * or the mean face (if defined).
     * @param i the first index
     * @param j the second index
     * @param pairwiseVisitor distance visitor
     * @param meanFaceVisitor mean face visitor
     */
    protected void setDistValues(int i, int j, MeshDistanceVisitor pairwiseVisitor, MeshDistanceVisitor meanFaceVisitor) {
        if (meanFaceVisitor == null) { // no cropping
            //setDistSimilarity(i, j, pairwiseVisitor.getStats().getMax());
            setDistSimilarity(i, j, pairwiseVisitor.getDistancesOfVisitedFacets().getDistanceStats().getAverage());
            setDistDeviation(i, j, pairwiseVisitor.getDistancesOfVisitedFacets().getSampleStandardDeviation());
            return;
        }

        // Auto crop to the average face:
        Map<MeshFacet, List<Double>> pairwiseMap = pairwiseVisitor.getDistancesOfVisitedFacets().distancesAsMap();
        Map<MeshFacet, List<Double>> meanFaceMap = meanFaceVisitor.getDistancesOfVisitedFacets().distancesAsMap();
        List<Double> values = new ArrayList<>();
        
        for (MeshFacet facet: pairwiseMap.keySet()) {
            for (int k = 0; k < facet.getNumberOfVertices(); k++) {
                if (Double.isFinite(meanFaceMap.get(facet).get(k)) &&
                        Double.isFinite(pairwiseMap.get(facet).get(k))) {
                    values.add(pairwiseMap.get(facet).get(k));
                }
            }
        }
        
        double avg = values.stream()
                .mapToDouble(Double::doubleValue)
                .average()
                .orElse(0.0);
        
        double sum = values.stream()
                .mapToDouble(Double::doubleValue)
                .map(dist -> (dist - avg) * (dist- avg))
                .sum();

        double dev = Math.sqrt((1.0 / (values.size()-1)) * sum);
        
        setDistSimilarity(i, j, avg);
        //setDistSimilarity(i, j, values.stream()
        //        .mapToDouble(Double::doubleValue)
        //        .max()
        //        .orElse(Double.POSITIVE_INFINITY)
        //);
        setDistDeviation(i, j, dev);
    }
}
