package cz.fidentis.analyst.gui.task.batch.distance;

import com.jogamp.opengl.GLContext;
import cz.fidentis.analyst.Logger;
import cz.fidentis.analyst.data.face.HumanFace;
import cz.fidentis.analyst.data.face.HumanFaceMemoryManager;
import cz.fidentis.analyst.engines.distance.MeshDistanceConfig;
import cz.fidentis.analyst.engines.distance.MeshDistanceServices;
import cz.fidentis.analyst.engines.distance.measurement.MeshDistances;
import cz.fidentis.analyst.engines.distance.measurement.DistanceRecord;
import cz.fidentis.analyst.engines.face.FaceStateServices;
import cz.fidentis.analyst.gui.elements.ProgressDialog;
import cz.fidentis.analyst.gui.task.batch.Stopwatch;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.logging.Level;

/**
 * A task that computes similarity of a set of faces by casting rays
 * in the diction of template face normals.
 * <p>
 * The computation is accelerated by using multiple CPU cores concurrently.
 * The exact computation parameters are taken from the {@code BatchDistancePanel}.
 * </p>
  * 
 * @author Radek Oslejsek
 */
public class RayCastingDistanceTask extends BatchDistanceTask {
    
    private final Stopwatch totalTime = new Stopwatch("Total computation time:\t\t");
    private final Stopwatch distComputationTime = new Stopwatch("Distance computation time:\t");
    private final Stopwatch loadTime = new Stopwatch("Disk access time:\t\t");
    private final Stopwatch octreeConstructionTime = new Stopwatch("Octrees construction time:\t\t");
    private final Stopwatch finalDistComputationTime = new Stopwatch("Distance finalization time:\t");
    private final GLContext context;
    
    /**
     * Constructor.
     * 
     * @param progressDialog A window that show the progress of the computation. Must not be {@code null}
     * @param controlPanel A control panel with computation parameters. Must not be {@code null}
     * @param context Active OpenGL context on which makeCurrent() can be called, if not {@code null}, then GPU will be used.
     *                Otherwise, regular CPU implementation is used.
     */
    public RayCastingDistanceTask(ProgressDialog<Void, Integer> progressDialog, BatchDistancePanel controlPanel, GLContext context) {
        super(progressDialog, controlPanel);
        this.context = context;
    }
    
    @Override
    protected Void doInBackground() throws Exception {
        HumanFaceMemoryManager factory = getControlPanel().getFacesProxyDecorator().getFactory();
        factory.setReuseDumpFile(true); // it's safe because no changes are made to models
        factory.setStrategy(HumanFaceMemoryManager.Strategy.MRU); // keep first X faces in the memory

        totalTime.start();

        // We don't need to reload the initFace periodically for two reasons:
        //   - It is never dumped from memory to disk because we use MRU
        //   - Even if dumped, the face keeps in the mempry until we hold the pointer to it
        loadTime.start();
        HumanFace gaugeFace = getSelectedFace();
        loadTime.stop();
        
        // Cache of distances of individual vertices
        //   distCache.get(i) = i-th face
        //   distCache.get(i).get(j) = distance of j-th vertex of i-th face
        DistCache distCache = new DistCache();

        int numFaces = getControlPanel().getFacesProxyDecorator().getFaceFiles().size();
        for (int i = 0; i < numFaces; i++) {

            if (isCancelled()) { // the user canceled the process
                return null;
            }
            
            loadTime.start();
            HumanFace face = getFaceFromFaceFactory(i);
            loadTime.stop();

            if(context == null) {
                octreeConstructionTime.start();
                FaceStateServices.updateOctree(face, FaceStateServices.Mode.COMPUTE_IF_ABSENT);
                octreeConstructionTime.stop();
            }

            distComputationTime.start();
            MeshDistances dist = MeshDistanceServices.measure(
                    gaugeFace.getMeshModel(),
                    context == null ? new MeshDistanceConfig(MeshDistanceConfig.Method.RAY_CASTING, face.getOctree(), true)
                                    : new MeshDistanceConfig(MeshDistanceConfig.Method.RAY_CASTING_GPU, face.getMeshModel().getFacets(), context, true, false));
            distCache.addValuesOfFace(dist);
            distComputationTime.stop();
                
            // Memory is more valuable. And it's probable that we will not need
            // the octrees anymore. If yes, then they are quickly re-computed.
            FaceStateServices.updateOctree(face, FaceStateServices.Mode.DELETE);

            // update progress bar
            getProgressDialog().setValue((int) Math.round(100.0 * (i+1) / numFaces));
        }
        
        finalDistComputationTime.start();
        finalizeDistance(distCache);
        finalDistComputationTime.stop();
        
        totalTime.stop();

        printTimeStats();

        return null;
    }
    
    protected void printTimeStats() {
        Logger.print(distComputationTime.toString());
        Logger.print(finalDistComputationTime.toString());
        Logger.print(loadTime.toString());
        Logger.print(octreeConstructionTime.toString());
        Logger.print(totalTime.toString());
    }
    
    private void finalizeDistance(DistCache distCache) {
        ExecutorService executor = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
        final List<Future<Void>> results = new ArrayList<>();
        
        for (int i = 0; i < distCache.getNumFaces(); i++) {
            for (int j = i; j < distCache.getNumFaces(); j++) {
                //assert(distCache.get(i).size() == distCache.get(i).size()); // should never happen
                
                final int fi = i;
                final int fj = j;
                
                // Compute average HD for the pair of faces in a separate threat:
                results.add(executor.submit(() -> {
                    if (fi == fj) {
                        setDistSimilarity(fi, fj, 0.0);
                        setDistSimilarity(fj, fi, 0.0);
                        setDistDeviation(fi, fj, 0.0);
                        setDistDeviation(fj, fi, 0.0);
                        return null;
                    }

                    double sum = 0.0;
                    int counter = 0;
                    for (int k = 0; k < distCache.getNunFaceVertices(fi); k++) {
                        double d = distCache.getDistance(fi, fj, k);
                        if (Double.isFinite(d)) {
                            sum += d;
                            counter++;
                        }
                    }
                    double dev = Math.sqrt((1.0 / (counter-1)) * sum);
                    setDistSimilarity(fi, fj, sum / counter);
                    setDistSimilarity(fj, fi, sum / counter);
                    setDistDeviation(fi, fj, dev);
                    setDistDeviation(fj, fi, dev);
                    return null;
                }));
            }
        }
        
        executor.shutdown();
        while (!executor.isTerminated()){}
        try {
            for (Future<Void> res: results) {
                res.get(); // waits until all computations are finished
            }
        } catch (final InterruptedException | ExecutionException ex) {
            java.util.logging.Logger.getLogger(NearestNeighborsDistanceTask.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
    
    
    /**
     * Cache of distances from the average face to other faces.
     * 
     * @author Radek Oslejsek
     */
    private class DistCache {
         private List<List<Double>> cache = new ArrayList<>();
         
        public void addValuesOfFace(MeshDistances distances) {
            if (distances == null) {
                add(null);
            } else { // Store relative distances of vertices to the cache
                add(distances.stream()
                        .mapToDouble(DistanceRecord::getDistance)
                        .boxed()
                        .toList());
            }
        }
        
        public double getDistance(int faceIndex1, int faceIndex2, int vertIndex) {
            double d1 = getCache().get(faceIndex1).get(vertIndex);
            double d2 = getCache().get(faceIndex2).get(vertIndex);
            return (Double.isFinite(d1) && Double.isFinite(d2))
                    ? Math.abs(d1 - d2)
                    : Double.NaN;
        }

         /**
          * Returns the number of cached faces
          * @return the number of cached faces
          */
         public int getNumFaces() {
             return cache.size();
         }
         
         /**
          * Returns the number of cached vertices of the given faces
          * @return the number of cached vertices of the given faces
          */
         public int getNunFaceVertices(int faceIndex) {
            return cache.get(faceIndex).size();
         }
         
         protected List<List<Double>> getCache() {
             return Collections.unmodifiableList(cache);
         }
         
         protected void add(List<Double> list) {
             cache.add(list);
         }
    }
    
}
