package cz.fidentis.analyst.gui.task.batch.registration;

import cz.fidentis.analyst.Logger;
import cz.fidentis.analyst.canvas.Canvas;
import cz.fidentis.analyst.data.face.HumanFace;
import cz.fidentis.analyst.data.face.HumanFaceFactory;
import cz.fidentis.analyst.data.face.HumanFaceMemoryManager;
import cz.fidentis.analyst.data.mesh.MeshIO;
import cz.fidentis.analyst.data.mesh.MeshModel;
import cz.fidentis.analyst.drawables.DrawableFace;
import cz.fidentis.analyst.engines.avgmesh.AvgMeshConfig;
import cz.fidentis.analyst.engines.avgmesh.AvgMeshVisitor;
import cz.fidentis.analyst.engines.icp.IcpConfig;
import cz.fidentis.analyst.engines.sampling.PointSamplingConfig;
import cz.fidentis.analyst.engines.face.FaceRegistrationServices;
import cz.fidentis.analyst.engines.face.FaceStateServices;
import cz.fidentis.analyst.gui.elements.ProgressDialog;
import cz.fidentis.analyst.gui.task.batch.Stopwatch;

import javax.swing.*;
import java.io.File;
import java.io.IOException;
import java.util.List;

/**
 * A concurrent task that registers multiple faces to selected template face. Simultaneously, it computes
 * average face. The exact computation parameters are taken from the
 * {@link BatchRegistrationPanel}.
 *
 * @author Radek Oslejsek
 */
public class BatchRegistrationTask extends SwingWorker<MeshModel, HumanFace> {

    private static final double ICP_ERROR = 0.05;
    private static final int ICP_ITERATIONS = 100;
    private static final int ICP_AUTO_CROP_SINCE = 1;

    private HumanFace avgFace;
    private final ProgressDialog<MeshModel,HumanFace> progressDialog;
    private final BatchRegistrationPanel controlPanel;
    private final Canvas canvas;

    private int faceSceneSlot = -1; // scene slot to show the transformed face

    private final Stopwatch totalTime = new Stopwatch("Total computation time:\t");
    private final Stopwatch avgFaceComputationTime = new Stopwatch("AVG face computation time:\t");
    private final Stopwatch registrationTime = new Stopwatch("Registration time:\t");
    private final Stopwatch loadTime = new Stopwatch("Disk access time:\t");
    private final Stopwatch kdTreeConstructionTime = new Stopwatch("KD trees construction time:\t");
    private final Stopwatch octreeConstructionTime = new Stopwatch("Octree construction time:\t");

    /**
     * Constructor.
     *
     * @param progressDialog A window that show the progress of the computation. Must not be {@code null}
     * @param controlPanel A control panel with computation parameters. Must not be {@code null}
     * @param canvas Canvas with a 3D scene.
     */
    public BatchRegistrationTask(ProgressDialog<MeshModel,HumanFace> progressDialog, BatchRegistrationPanel controlPanel, Canvas canvas) {
        this.progressDialog = progressDialog;
        this.controlPanel = controlPanel;
        this.canvas = canvas;
    }

    @Override
    protected MeshModel doInBackground() throws Exception {
        // Get computational parameters:
        String regStrategy = controlPanel.getRegistrationStrategy();
        String avgFaceStrategy = controlPanel.getAvgFaceStrategy();
        boolean doRegistration = regStrategy.equals(BatchRegistrationPanel.REGISTRATION_GPA) ||
                regStrategy.equals(BatchRegistrationPanel.REGISTRATION_ICP);
        boolean computeAvgFace = avgFaceStrategy.equals(BatchRegistrationPanel.AVG_FACE_NEAREST_NEIGHBOURS) ||
                avgFaceStrategy.equals(BatchRegistrationPanel.AVG_FACE_RAY_CASTING);

        if (!(doRegistration || computeAvgFace)) { // nothing to do
            return null;
        }

        setHumanFaceFactoryParams();

        totalTime.start();

        // Get initial (template) face. Either the average face is chosen if exists (we want to improve previous iteration),
        // or the selected face is chosen (the first iteration).
        // We don't need to reload the initFace periodically for two reasons:
        //   - It is never dumped from memory to disk because we use MRU
        //   - Even if dumped, the face keeps in the memory until we hold the pointer to it
        HumanFace initFace = getTemplateFace();

        // Iterate through all faces to register them and/or compute new average face.
        // Process all faces, including the selected one, because the template can vary.
        AvgMeshVisitor avgFaceConstructor = null;
        int numFaces = controlPanel.getFacesProxyDecorator().getFaceFiles().size();
        for (int i = 0; i < numFaces; i++) {
            if (isCancelled()) {
                return null;
            }

            loadTime.start();
            HumanFace superimposedFace = getFaceFromFaceFactory(i); // Load face to be processed and possibly transformed
            loadTime.stop();

            if (doRegistration) { // superimpose superimposedFace to initFace
                register(regStrategy, initFace, superimposedFace);
            }

            if (computeAvgFace) { // update the average face
                avgFaceConstructor = (avgFaceStrategy.equals(BatchRegistrationPanel.AVG_FACE_NEAREST_NEIGHBOURS))
                        ? computeAvgFaceNN(initFace, superimposedFace, avgFaceConstructor, doRegistration)
                        : computeAvgFaceRT(initFace, superimposedFace, avgFaceConstructor, doRegistration);
            }

            publish(superimposedFace);       // update progress bar and possibly render the transformed face
            progressDialog.setValue((int) Math.round(100.0 * (i + 1.0) / numFaces)); // set progress value
        }

        if (avgFaceConstructor != null) { // create new human face from the average face geometry
            avgFace = createAvgFace(avgFaceConstructor.getAveragedMeshModel());
        }

        totalTime.stop();
        printTimeStats();

        return (avgFace == null) ? null : avgFace.getMeshModel();
    }

    @Override
    protected void done() {
        progressDialog.dispose(); // close progress bar
        if (isCancelled()) {
            avgFace = null;
        }
    }

    @Override
    protected void process(List<HumanFace> chunks) {
        chunks.forEach(f -> {
            if (isCancelled()) {
                return;
            }
            if (faceSceneSlot == -1) {
                faceSceneSlot = canvas.getScene().getFreeSlotForFace();
            }
            canvas.getScene().setDrawableFace(faceSceneSlot, f);
            canvas.getScene().getDrawableFace(faceSceneSlot).setTransparency(0.5f);
            canvas.getScene().getDrawableFace(faceSceneSlot).setColor(DrawableFace.SKIN_COLOR_SECONDARY);
            canvas.renderScene();
        });
    }

    public HumanFace getAverageFace() {
        return this.avgFace;
    }

    protected void printTimeStats() {
        Logger.print(avgFaceComputationTime.toString());
        Logger.print(registrationTime.toString());
        Logger.print(loadTime.toString());
        Logger.print(kdTreeConstructionTime.toString());
        Logger.print(octreeConstructionTime.toString());
        Logger.print(totalTime.toString());
    }
    
    protected void register(String regStrategy, HumanFace initFace, HumanFace superimposedFace) {
        registrationTime.start();
        switch (regStrategy) {
            case BatchRegistrationPanel.REGISTRATION_ICP -> { // ICP registration - the face is transformed!
                int subsampling = controlPanel.getIcpUndersampling();
                PointSamplingConfig sampling = (subsampling == 0)
                        ? new PointSamplingConfig(PointSamplingConfig.Method.NO_SAMPLING, subsampling)
                        : new PointSamplingConfig(PointSamplingConfig.Method.RANDOM, subsampling);
                FaceRegistrationServices.alignMeshes(
                        superimposedFace, // is transformed
                        new IcpConfig(
                                initFace.getMeshModel(),
                                ICP_ITERATIONS,
                                controlPanel.scaleIcp(),
                                ICP_ERROR,
                                sampling,
                                ICP_AUTO_CROP_SINCE));
            }
            case BatchRegistrationPanel.REGISTRATION_GPA -> // GPA registration - the face is transformed!
                FaceRegistrationServices.alignFeaturePoints(
                        initFace,
                        superimposedFace, // is transformed
                        controlPanel.scaleIcp());
            default -> throw new IllegalStateException("Unexpected value: " + regStrategy);
        }
        registrationTime.stop();
    }

    protected AvgMeshVisitor computeAvgFaceRT(HumanFace initFace, HumanFace superimposedFace, AvgMeshVisitor avgFaceConstructor, boolean doRegistration) {
        // If the face was moved by registration, then the spatial ordering structure was removed => create it
        // If no transformation was made, then you can use old structure, if exists
        octreeConstructionTime.start();
        FaceStateServices.updateOctree(
                superimposedFace,
                doRegistration ? FaceStateServices.Mode.COMPUTE_ALWAYS : FaceStateServices.Mode.COMPUTE_IF_ABSENT
        );
        octreeConstructionTime.stop();

        avgFaceComputationTime.start();
        AvgMeshVisitor ret = (avgFaceConstructor == null)
                ? (new AvgMeshConfig(initFace.getMeshModel())).getRayCastingVisitor()
                : avgFaceConstructor;
        superimposedFace.getOctree().accept(ret);
        avgFaceComputationTime.stop();
        
        return ret;
    }

    protected AvgMeshVisitor computeAvgFaceNN(HumanFace initFace, HumanFace superimposedFace, AvgMeshVisitor avgFaceConstructor, boolean doRegistration) {
        // If the face was moved by registration, then the spatial ordering structure was removed => create it
        // If no transformation was made, then you can use old structure, if exists
        kdTreeConstructionTime.start();
        FaceStateServices.updateKdTree(
                superimposedFace,
                doRegistration ? FaceStateServices.Mode.COMPUTE_ALWAYS : FaceStateServices.Mode.COMPUTE_IF_ABSENT
        );
        kdTreeConstructionTime.stop();

        avgFaceComputationTime.start();
        AvgMeshVisitor ret = (avgFaceConstructor == null)
                ? (new AvgMeshConfig(initFace.getMeshModel())).getNearestNeighborsVisitor()
                : avgFaceConstructor;
        superimposedFace.getKdTree().accept(ret);
        avgFaceComputationTime.stop();

        return ret;
    }

    protected HumanFace createAvgFace(MeshModel avgMesh) throws IOException {
        File tempFile = File.createTempFile(this.getClass().getSimpleName(), ".obj");
        tempFile.deleteOnExit();

        HumanFace newAvgFace = HumanFaceFactory.create(avgMesh, tempFile.getCanonicalPath());
        try {
            MeshIO.exportMeshModel(newAvgFace.getMeshModel(), tempFile);
        } catch (IOException ex) {
            Logger.print(ex.toString());
        }

        return newAvgFace;
    }

    protected HumanFace getFaceFromFaceFactory(int faceIndex) {
        String faceId = controlPanel
                .getFacesProxyDecorator()
                .getFactory()
                .loadFace(controlPanel.getFacesProxyDecorator().getFaceFiles().get(faceIndex));
        return controlPanel.getFacesProxyDecorator().getFactory().getFace(faceId);
    }

    protected void setHumanFaceFactoryParams() {
        HumanFaceMemoryManager factory = controlPanel.getFacesProxyDecorator().getFactory();
        factory.setReuseDumpFile(false); // we can't use this optimization because changes are made in models
        factory.setStrategy(HumanFaceMemoryManager.Strategy.MRU); // keep first X faces in the memory
    }

    protected HumanFace getTemplateFace() {
        loadTime.start();
        HumanFace ret = controlPanel.getFacesProxyDecorator().haveAvgFace()
                ? controlPanel.getFacesProxyDecorator().getAvgFace()
                : controlPanel.getSelectedFace();
        loadTime.stop();
        return ret;
    }
}
