package cz.fidentis.analyst.gui.task.curvature;

import cz.fidentis.analyst.canvas.Canvas;
import cz.fidentis.analyst.engines.face.FaceStateServices;
import cz.fidentis.analyst.gui.task.ControlPanelAction;
import cz.fidentis.analyst.data.mesh.MeshFacet;
import cz.fidentis.analyst.project.FacesProxy;
import cz.fidentis.analyst.data.shapes.HeatMap3D;

import javax.swing.*;
import java.awt.event.ActionEvent;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Action listener for the curvature computation.
 * 
 * @author Radek Oslejsek
 */
public class CurvatureAction extends ControlPanelAction<CurvaturePanel> {
    
    /*
     * Attributes handling the state
     */
    private String curvatureType = CurvaturePanel.GAUSSIAN_CURVATURE;

    /**
     * Constructor.
     * A new {@code CurvaturePanel} is instantiated and added to the {@code topControlPane}
     * 
     * @param canvas OpenGL canvas
     * @param faces Faces processed by current analytical task
     * @param topControlPane A top component when a new control panel is placed
     */
    public CurvatureAction(Canvas canvas, FacesProxy faces, JTabbedPane topControlPane) {
        super(canvas, faces, topControlPane);
        setControlPanel(new CurvaturePanel(this));
        
        setShowHideCode(
                () -> { // SHOW
                    // provide a code that runs when the panel is focused (selected)
                    setHeatmap();
                },
                () -> { // HIDE
                    // provide a code that runs when the panel is closed (hidden)
                    getPrimaryDrawableFace().setHeatMap(null);
                }
        );
    }

    @Override
    public void actionPerformed(ActionEvent ae) {
        String action = ae.getActionCommand();
        
        switch (action) {
            case CurvaturePanel.ACTION_COMMAND_SET_CURVATURE_TYPE:
                this.curvatureType = (String) ((JComboBox<?>) ae.getSource()).getSelectedItem();
                setHeatmap();
                break;
            default:
                // to nothing
        }
        
        renderScene();
    }
    
    protected void setHeatmap() {
        FaceStateServices.updateCurvature(getPrimaryDrawableFace().getHumanFace(), FaceStateServices.Mode.COMPUTE_IF_ABSENT);
        Map<MeshFacet, List<Double>> heatmapValues = new HashMap<>();
        for (MeshFacet facet: getPrimaryDrawableFace().getModel().getFacets()) {
            List<Double> values = facet.getVertices().stream()
                    .map(mp -> switch (this.curvatureType) {
                        case CurvaturePanel.GAUSSIAN_CURVATURE -> mp.getCurvature().gaussian();
                        case CurvaturePanel.MEAN_CURVATURE -> mp.getCurvature().mean();
                        case CurvaturePanel.MIN_CURVATURE -> mp.getCurvature().minPrincipal();
                        case CurvaturePanel.MAX_CURVATURE -> mp.getCurvature().maxPrincipal();
                        default -> throw new UnsupportedOperationException(curvatureType);
                    })
                    .toList();
            heatmapValues.put(facet, values);
        }
        getPrimaryDrawableFace().setHeatMap(new HeatMap3D(heatmapValues, null));
    }
}
