package cz.fidentis.analyst.data.shapes;

import cz.fidentis.analyst.data.mesh.MeshFacet;
import java.util.ArrayList;
import java.util.List;
import java.util.Collections;
import javax.vecmath.Point3d;

/**
 * Class representing the Interactive Mask, it is defined as a list of points
 * lines between them and bounding cylinders.
 * 
 * @author Mario Chromik
 */
public class SurfaceMask {
    
    public static final int CYLIDER_RADIUS = 10;
    
    private final List<MaskPoint> points = new ArrayList<>();
    private final List<Cylinder> cylinders = new ArrayList<>();
    
    /**
     * Returns {@code true} if the mast has no point defined so far.
     * 
     * @return {@code true} if the mast has no point defined so far.
     */
    public boolean isEmpty() {
        return points.isEmpty();
    }
    
    /**
     * Adds point of Interactive Mask
     * @param surfacePoint a point lying on the surface
     * @param facet facet of the surface
     */
    public void addPoint(Point3d surfacePoint, MeshFacet facet) {
        if (points.isEmpty()) {
           points.add(new MaskPoint(surfacePoint, facet));
           return;
        }
        
        if (points.size() == 1) {
           points.add(new MaskPoint(surfacePoint, facet));
           cylinders.add(new Cylinder(points.get(0).getPosition(), surfacePoint, CYLIDER_RADIUS));
           return;
        }
        
        // add to the middle
        for (int i = 0; i < points.size() - 1; i++) {
            MaskPoint p1 = points.get(i);
            MaskPoint p2 = points.get(i + 1); 
            
            Cylinder cylinder = new Cylinder(p1.position, p2.position, CYLIDER_RADIUS);
            
            // if Bounding cyllinder contains the new point, add it in between 
            // the two points from which the bounding cyllinder has been constructed
            if (cylinder.containsPoint(surfacePoint)) {
                points.add(i + 1, new MaskPoint(surfacePoint, facet));
                cylinders.set(i, new Cylinder(points.get(i).getPosition(), surfacePoint, CYLIDER_RADIUS));
                cylinders.add(i + 1, new Cylinder(surfacePoint, points.get(i+1).getPosition(), CYLIDER_RADIUS));
                return;
            }
        }
        
        // add to the end
        points.add(new MaskPoint(surfacePoint, facet));
        cylinders.add(new Cylinder(points.get(points.size()-1).getPosition(), surfacePoint, CYLIDER_RADIUS));
    }
    
    /**
     * to be implemented
     * @param point 
     */
    public void removePoint(Point3d point) {
        //TODO
    }
    
    /**
     * to be implemented
     * @param origin
     * @param target 
     */
    public void movePoint(Point3d origin, Point3d target) {
        //TODO
    }
    
    /**
     * get points of interactive mask
     * @return point
     */
    public List<MaskPoint> getPoints() {
        return Collections.unmodifiableList(points);
    }
    
    /**
     * gets cylinders
     * @return bounding cylinders
     */
    public List<Cylinder> getCyllinders() {
        return Collections.unmodifiableList(cylinders);
    }
    
    /**
     * Class representing a single point of Interactive Mask.
     * 
     * @author Mario Chromik
     */
    public static final class MaskPoint {
    
        private final Point3d position;
        private final MeshFacet facet;
        
        /**
         * Constructor.
         * 
         * @param pos position of an Interactive Mask point in 3D space
         * @param fac facet of IM point
         */
        private MaskPoint(Point3d pos, MeshFacet fac) {
            position = pos;
            facet = fac;
        }
    
        public Point3d getPosition() {
            return position;
        }
        
        public MeshFacet getFacet() {
            return facet;
        }
    }
}