package cz.fidentis.analyst.engines.symmetry.impl;

import cz.fidentis.analyst.data.mesh.MeshFacet;
import cz.fidentis.analyst.data.mesh.MeshPoint;
import cz.fidentis.analyst.data.shapes.Plane;
import cz.fidentis.analyst.engines.bbox.BoundingBoxConfig;
import cz.fidentis.analyst.engines.bbox.BoundingBoxVisitor;
import cz.fidentis.analyst.engines.curvature.CurvatureConfig;
import cz.fidentis.analyst.engines.curvature.CurvatureServices;
import cz.fidentis.analyst.engines.sampling.PointSamplingConfig;
import cz.fidentis.analyst.engines.sampling.PointSamplingVisitor;
import cz.fidentis.analyst.engines.symmetry.SymmetryVisitor;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.*;
import java.util.logging.Level;

/**
 * A old implementation of the symmetry plane estimator taken from the old FIDENTIS.
 * Conceptually, it is similar to the {@link cz.fidentis.analyst.engines.symmetry.impl.SymmetryVisitorRobustMesh}.
 * It also uses Gaussian curvature and normal vectors to prune candidate planes 
 * and measure their quality. But some additional criteria are used as well. 
 * Also, the calculation of "votes" differs.
 * This implementation has the following properties:
 * <ul>
 * <li>Fast. A bit faster than the new robust algorithms of {@link cz.fidentis.analyst.engines.symmetry.impl.SymmetryVisitorRobust}
 * and much faster that the {@link cz.fidentis.analyst.engines.symmetry.impl.SymmetryVisitorRobustMesh}.</li>
 * <li>Best candidate planes are chosen based on the similarity of Gauss curvatures 
 * on both sides of the plane, inverse direction of normal vectors, 
 * and the distance of plane from centroid.</li>
 * <li>Because the space of candidate planes is roughly approximated (several true-false rules), 
 * the final symmetry plane is computed by averaging best candidates (candidates with the most votes).</li>
 * <li>Best results are achieved with Uniform Grid sampling with 200-300 points and averaging turn on.</li>
 * <li>Should return reasonable results also for incomplete faces (was not tested).</li>
 * </ul>
 * <p>
 * This visitor <b>is not thread-safe</b>, i.e., a single instance of the visitor 
 * cannot be used to inspect multiple meshes simultaneously  (sequential inspection is okay). 
 * It because the underlying visitors may not be thread-safe.
 * </p>
 * <p>
 * The main symmetry plane computation is performed by the {@link #getSymmetryPlane()} method,
 * not the {@link #visitMeshFacet(MeshFacet)}.
 * </p>
 * 
 * @author Natalia Bebjakova
 * @author Radek Oslejsek
 */
public class SymmetryVisitorMesh extends SymmetryVisitor {
    
    public static final double MAX_REL_DISTANCE = 0.01;
    
    /**
     * If {@code true}, then the final symmetry plane is computed by averaging the best candidates.
     * If {@code false}, then the random top candidate is selected.
     */
    public static final boolean AVERAGING = true;

    private final PointSamplingVisitor pointSamplingVisitor;
    private final BoundingBoxVisitor bbVisitor = (new BoundingBoxConfig()).getVisitor();
    
    private CandidatePlaneMesh symmetryPlane;
    
    /**
     * Constructor.
     * 
     * @param samplingStrategy Sub-sampling strategy. Must not be {@code null}
     * @throws IllegalArgumentException if some input parameter is missing
     */
    public SymmetryVisitorMesh(PointSamplingConfig samplingStrategy) {
        if (samplingStrategy == null) {
            throw new IllegalArgumentException("samplingStrategy");
        }
        this.pointSamplingVisitor = samplingStrategy.getVisitor();
    }
    
    @Override
    public boolean isThreadSafe() {
        return false;
    }
    
    @Override
    public void visitMeshFacet(MeshFacet facet) {
        // We need vertex normals for the plane computation
        synchronized (this) {
            if (!facet.hasVertexNormals()) {
                facet.calculateVertexNormals();
            }
        }
        
        // We need Gaussian curvatures
        synchronized (this) {
            if (facet.getVertex(0).getCurvature() == null) {
                CurvatureServices.computeAndSet(facet, new CurvatureConfig());
            }
        }
        
        // We want to sub-sample the mesh
        pointSamplingVisitor.visitMeshFacet(facet);
        
        // We need bounding box to compute max distance
        bbVisitor.visitMeshFacet(facet);
    }
    
    @Override
    public Plane getSymmetryPlane() {
        return getSymmetryPlane(true);
    }

    @Override
    public void dispose(){
        pointSamplingVisitor.dispose();
    }
    
    /**
     * Computes and returns the symmetry plane. The plane is computed only once, 
     * then the same instance is returned.
     * 
     * @param concurrently If {@code true}, then parallel computation is used utilizing all CPU cores.
     * @return the symmetry plane or {@code null}
     */
    public Plane getSymmetryPlane(boolean concurrently) {
        if (symmetryPlane == null) {
            this.calculateSymmetryPlane(concurrently);
        }
        return symmetryPlane.getPlane();
    }
    
    /**
     * Calculates the symmetry plane.
     * @param concurrently If {@code true}, then parallel computation is used utilizing all CPU cores.
     * Otherwise, the computation is sequential.
     */
    protected void calculateSymmetryPlane(boolean concurrently) {
        final List<CandidatePlaneMesh> planes = new ArrayList<>();
        final double maxDistance = bbVisitor.getBoundingBox().diagonalLength() * MAX_REL_DISTANCE;
        
        List<MeshPoint> sigVertices = pointSamplingVisitor.getSamples();
        SymmetryCache cache = new SymmetryCache(sigVertices);
        
        /*
         * Sequential computation
         */
        if (!concurrently) {
            int maxVotes = 0;
            for (int i = 0; i < sigVertices.size(); i++) {
                for (int j = 0; j < sigVertices.size(); j++) {
                    CandidatePlaneMesh newPlane;
                    try {
                        newPlane = new CandidatePlaneMesh(sigVertices, cache, i, j, maxDistance);
                    } catch(UnsupportedOperationException ex) {
                        continue;
                    }
                    maxVotes = checkAndUpdatePlanes(planes, newPlane, maxVotes);
                }
            }
            setSymmetryPlane(planes);
            return;
        }
        
        /*
         * Concurrent computation
         */
        
        // Initiate structure for concurrent computation:
        ExecutorService executor = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
        final List<Future<CandidatePlaneMesh>> results = new ArrayList<>(sigVertices.size() * sigVertices.size());
        
        // Compute candidate planes concurrently:
        for (int i = 0; i < sigVertices.size(); i++) {
            for (int j = 0; j < sigVertices.size(); j++) {
                int finalI = i;
                int finalJ = j;
                
                Callable<CandidatePlaneMesh> worker = () -> {
                    try {
                        return new CandidatePlaneMesh(sigVertices, cache, finalI, finalJ, maxDistance);
                    } catch(UnsupportedOperationException ex) {
                        return null;
                    }
                };
                
                results.add(executor.submit(worker));
            }
        }
        
        // Wait until all symmetry planes are computed:
        executor.shutdown();
        while (!executor.isTerminated()){}
        
        // Get the best-fitting planes:
        try {
            int maxVotes = 0;
            for (Future<CandidatePlaneMesh> res: results) {
                CandidatePlaneMesh newPlane = res.get();
                if (newPlane != null) {
                    maxVotes = checkAndUpdatePlanes(planes, newPlane, maxVotes);
                }
            }
        } catch (final InterruptedException | ExecutionException ex) {
            java.util.logging.Logger.getLogger(SymmetryVisitorMesh.class.getName()).log(Level.SEVERE, null, ex);
        }
        
        setSymmetryPlane(planes);
    }
    
    protected int checkAndUpdatePlanes(List<CandidatePlaneMesh> planes, CandidatePlaneMesh newPlane, int maxVotes) {
        if (newPlane.getVotes() > maxVotes) {
            planes.clear();
            planes.add(newPlane);
            return newPlane.getVotes();
        } else if (newPlane.getVotes() == maxVotes) {
            planes.add(newPlane);
        }
        return maxVotes;
    }

    protected void setSymmetryPlane(List<CandidatePlaneMesh> planes) {
        if (AVERAGING && !planes.isEmpty()) {
            symmetryPlane = new CandidatePlaneMesh(new Plane(planes.stream().map(cp -> cp.getPlane()).toList()));
        } else {
            symmetryPlane = planes.isEmpty() ? null : new CandidatePlaneMesh(planes.get(0).getPlane());
        }
    }
    
}
