package cz.fidentis.analyst.project;

import cz.fidentis.analyst.data.face.HumanFace;
import cz.fidentis.analyst.data.face.HumanFaceFactory;
import cz.fidentis.analyst.data.face.HumanFaceMemoryManager;

import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.*;

/**
 * A proxy over the {@code HumanFaceFactory}. It handles faces processed by a single analytical task or project.
 * This class introduces lazy allocation of human faces in addition to automatic memory management provided
 * by the {@code HumanFaceFactory}.
 * 
 * @author Radek Oslejsek
 */
@SuppressWarnings("unchecked")
public class FacesProxy {
    
    /**
     * Types of files associated with a single human face
     * @author Radek Oslejsek
     */
    public enum FaceFileType {
        MESH, // .obj
        FEATURE_POINTS, // _landmarks.csv
        PREVIEW_IMAGE, // _preview.jpg
        INFO, // _info.csv
        TEXTURE // .jpg
    }
    
    /**
     * The factory instantiating human faces and automatically
     * swapping them onto the dist if the application runs out of memory.
     */
    private final HumanFaceMemoryManager factory = HumanFaceMemoryManager.create();
    
    /**
     * Paths to face files.
     */
    private final List<File> faceFiles = new ArrayList<>();
    
    /**
     * Constructor for multiple faces, e.g., 1:N or N:N batch processing
     * 
     * @param faceFiles faces of the task
     */
    public FacesProxy(List<File> faceFiles) {
        this.faceFiles.addAll(faceFiles);
    }
    
    /**
     * Constructor for a single face analysis
     * 
     * @param faceFile face of the task
     */
    public FacesProxy(File faceFile) {
        this(List.of(faceFile));
    }
    
    /**
     * Constructor for the analysis of two faces
     * 
     * @param priFaceFile primary face
     * @param secFaceFile secondary face
     */
    public FacesProxy(File priFaceFile, File secFaceFile) {
        this(List.of(priFaceFile, secFaceFile));
    }

    /**
     * Constructor for multiple faces, e.g., 1:N or N:N batch processing
     * 
     * @param faceFiles faces of the task
     */
    public FacesProxy(File... faceFiles) {
        this(Arrays.asList(faceFiles));
    }
    
    /**
     * Deserialization constructor.
     * It recovers faces from dump files
     * 
     * @param in Input stream
     * @throws IOException on IO error
     * @throws ClassNotFoundException of format error
     */
    public FacesProxy(ObjectInputStream in) throws IOException, ClassNotFoundException {
        this((List<File>) in.readObject());
        for (File ignored : faceFiles) {
            HumanFace face = HumanFaceFactory.create(in);
            if (face != null) {
                this.factory.addFace(face);
            }
        }
    }
    
    /**
     * Serializes current state.
     * 
     * @param out Output stream
     * @throws IOException on IO error
     */
    public void serialize(ObjectOutputStream out) throws IOException {
        out.writeObject(faceFiles);
        for (File file: faceFiles) {
            if (this.factory.isLoaded(file)) { // skip unloaded faces
                String faceId = factory.loadFace(file);
                out.writeObject(this.factory.getFace(faceId));
            } else {
                out.writeObject(null);
            }
        }
        out.flush();
    }
    
    /**
     * Returns the list of paths to faces.
     * @return the list of paths to faces.
     */
    public List<File> getFaceFiles() {
        return Collections.unmodifiableList(faceFiles);
    }
    
    /**
     * Returns the file of a given face.
     * 
     * @param index index of the face
     * @return the file of a given face or {@code null}
     */
    public File getFaceFile(int index) {
        if (index < 0 || index >= faceFiles.size()) {
            return null;
        }
        return faceFiles.get(index);
    }
    
    /**
     * Returns the file of a given face.
     * 
     * @param faceName name of the face
     * @return the file of a given face or {@code null}
     */
    public File getFaceFile(String faceName) {
        return getFaceFile(getFaceIndex(faceName));
    }
    
    /**
     * <b>Instantiates</b> (if not yet instantiated) and returns a human face.
     * 
     * @param index index of the face
     * @return human face or {@code null}
     */
    public HumanFace getFace(int index) {
        if (index < 0 || index >= faceFiles.size()) {
            return null;
        }
        String faceId = factory.loadFace(faceFiles.get(index));
        return factory.getFace(faceId);
    }
    
    /**
     * Returns index of the face described by its canonical name,
     * @param faceName Canonical name of the face
     * @return index or -1
     */
    public int getFaceIndex(String faceName) {
        if (faceName == null) {
            return -1;
        }
        for (int i = 0; i < this.faceFiles.size(); i++) {
            if (faceFiles.get(i).toPath().endsWith(faceName+".obj")) {
                return i;
            }
        }
        return -1;
    }
    
    /**
     * Adds a face to the set.
     * 
     * @param file path to the face OBJ file
     * @param unique If {@code true}, then avoid adding a single face twice
     * @return index of the added face, -1 the {@code file} is null, -2 if {@code unique}
     * is {@code true} and the face is already in the set.
     */
    public int addFace(File file, boolean unique) {
        if (file != null) {
            if (faceFiles.contains(file)) {
                return -2;
            }
            faceFiles.add(file);
            return faceFiles.size() - 1;
        } else {
            return -1;
        }
    }
    
    /**
     * Removes face from the proxy.
     * 
     * @param faceName Canonical name of the face
     * @return {@code false} of the face does not exist or was not loaded
     */
    public boolean removeFace(String faceName) {
        int index = this.getFaceIndex(faceName);
        if (index < 0) {
            return false;
        }
        
        File faceFile = getFaceFile(index);
        if (factory.isLoaded(faceFile)) { // in memeory or on disk
            String faceId = factory.loadFace(faceFile); // get reference
            factory.removeFace(faceId);
        } 
        faceFiles.remove(index);
        return true;
    }
    
    /**
     * Checks whether the face is already loaded in the memory.
     * 
     * @param index index of the face
     * @return {@code true} is the face is loaded.
     */
    public boolean isInMemory(int index) {
        if (index < 0 || index >= faceFiles.size()) {
            return false;
        }
        return factory.isLoaded(faceFiles.get(index));
    }
    
    /**
     * Returns short (canonical) name of the face.
     * In contrast to the {@code getFace(x).getShortName()}, 
     * this method does not need face instance allocated in the memory.
     * 
     * @param index index of the face
     * @return short name of the face
     */
    public String getFaceName(int index) {
        if (index < 0 || index >= faceFiles.size()) {
            return null;
        }
        return getCanonicalName(faceFiles.get(index));
    }
    
    /**
     * Returns path to the given face file or {@code null}.
     * 
     * @param index index of the face
     * @param type face file type
     * @return path or {@code null}
     */
    public String getFacePath(int index, FaceFileType type) {
        if (index < 0 || index >= faceFiles.size()) {
            return null;
        }
        
        String expectedFile = faceFiles.get(index).getPath();
        switch (type) {
            case MESH:
                return (new File(expectedFile).exists()) ? expectedFile : null;
            case FEATURE_POINTS:
                expectedFile = expectedFile.substring(0, expectedFile.lastIndexOf("."));
                expectedFile = expectedFile.concat("_landmarks.csv");
                return (new File(expectedFile).exists()) ? expectedFile : null;
            case PREVIEW_IMAGE:
                expectedFile = expectedFile.substring(0, expectedFile.lastIndexOf("."));
                expectedFile = expectedFile.concat("_preview.jpg");
                return (new File(expectedFile).exists()) ? expectedFile : null;
            case INFO:
                expectedFile = expectedFile.substring(0, expectedFile.lastIndexOf("."));
                expectedFile = expectedFile.concat("_info.csv");
                return (new File(expectedFile).exists()) ? expectedFile : null;
            case TEXTURE:
                expectedFile = expectedFile.substring(0, expectedFile.lastIndexOf("."));
                expectedFile = expectedFile.concat(".jpg");
                return (new File(expectedFile).exists()) ? expectedFile : null;
            default:
                return null;
        }
    }
    
    /**
     * Returns path to the face directory or {@code null}. 
     * @param index index of the face
     * @return path to face folder or {@code null}
     */
    public String getFaceDir(int index) {
        String dir = getFacePath(index, FaceFileType.MESH);
        if (dir == null) {
            return null;
        }
        int i = dir.lastIndexOf(File.separatorChar);
        return (i > 0) ? dir.substring(0, i) : dir;
    }
    
    /**
     * <b>Instantiates</b> (if not yet instantiated) and returns 
     * the first face from the list of faces. The first face is called primary.
     * 
     * @return the first (primary) face from the list of faces
     */
    public HumanFace getPrimaryFace() {
        return getFace(0);
    }
    
    /**
     * <b>Instantiates</b> (if not yet instantiated) and returns 
     * the second face from the list of faces. The second face is called secondary.
     * 
     * @return the second (secondary) face from the list of faces
     */
    public HumanFace getSecondaryFace() {
        return getFace(1);
    }
    
    /**
     * Returns the number of faces
     * @return the number of faces
     */
    public int getNumFaces() {
        return this.faceFiles.size();
    }
    
    /**
     * Returns the task name that encodes names of faces.
     * 
     * @return a text that encodes names of faces.
     */
    public String getTaskName() {
        if (faceFiles.isEmpty()) {
            return "";
        } else if (faceFiles.size() == 1) {
            return getCanonicalName(faceFiles.get(0));
        } else if (faceFiles.size() == 2) {
            return getCanonicalName(faceFiles.get(0)) 
                    + ":" 
                    + getCanonicalName(faceFiles.get(1));
        } else {
            return faceFiles.size() + " faces";
        }
    }
    
    /**
     * Returns name of the file without appendix and path
     * 
     * @param file path to the file
     * @return canonical name
     */
    public static String getCanonicalName(File file) {
        return file.toString().substring(
                file.toString().lastIndexOf(File.separatorChar) + 1, 
                file.toString().lastIndexOf('.')
        );
    }
    
    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof FacesProxy other)) {
            return false;
        }
        return faceFiles.equals(other.faceFiles);
    }

    @Override
    public int hashCode() {
        return Objects.hashCode(faceFiles);
    }

    public HumanFaceMemoryManager getFactory() {
        return factory;
    }
}
