package cz.fidentis.analyst.project;

import com.fasterxml.jackson.databind.ObjectMapper;
import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.util.prefs.Preferences;
import java.util.stream.Collectors;

/**
 * A FIDENTIS project.
 *
 * @author Matej Kovar
 * @author Radek Oslejsek
 */
public class Project {
    
    private boolean changed = false;
    
    /**
     * Project's faces
     */
    private final FacesProxy faces;
    
    /**
     * JSON (de)serialization object
     */
    private final ObjectMapper jacksonMapper = new ObjectMapper();
    
    /**
     * Path to the project directory
     */
    private File dir;
    
    /**
     * Name of the JSON file
     */
    private static final String PROJECT_FILE_NAME = "fidentis";
    
    /**
     * Default name of a new unsaved project
     */
    private static final String DEFAULT_PROJECT_NAME = "New Project";
    
    /**
     * A preferences key for storing path to the recent project
     */
    private static final String RECENT_PROJECT_PREF_KEY = "pathToMostRecentProject";
    
    /**
     * Constructor for a new empty project.
     */
    public Project() {
        this.faces = new FacesProxy();
    }
    
    /**
     * Re-initializes an existing project.
     * 
     * @param path Path to the project's dir
     * @throws java.io.IOException on IO error
     */
    public Project(File path) throws IOException {
        this.dir = path;
        File file = getProjectFile(dir);
        ProjectJsonStructure cfg = jacksonMapper.readValue(file, ProjectJsonStructure.class); 
        this.faces = new FacesProxy(cfg.getPaths().stream()
                .map(Path::toFile)
                .collect(Collectors.toList())
        );
        Preferences userPreferences = Preferences.userNodeForPackage(Project.class); 
        userPreferences.put(RECENT_PROJECT_PREF_KEY, file.getAbsolutePath());
    }
    
    /**
     * Returns a path to the JSON project file
     * 
     * @param dir Project's folder
     * @return path to the JSON project file
     */
    public static File getProjectFile(File dir) {
        return new File(dir, PROJECT_FILE_NAME + ".json");
    }
    
    /**
     * Returns a path to the last stored JSON project file
     * 
     * @return path to the last stored JSON project file or {@code null}
     */
    public static File getRecentProjectFile() {
        Preferences userPreferences = Preferences.userNodeForPackage(Project.class); 
        String path = userPreferences.get(RECENT_PROJECT_PREF_KEY, "");
        return ("".equals(path)) ? null : new File(path);
    }
    
    /**
     * Returns a path to the last stored project file folder
     * 
     * @return a path to the last stored project file folder or {@code null}
     */
    public static File getRecentProjectDir() {
        Preferences userPreferences = Preferences.userNodeForPackage(Project.class); 
        String path = userPreferences.get(RECENT_PROJECT_PREF_KEY, "");
        if ("".equals(path)) {
            return null;
        }
        return new File(path.substring(0, path.lastIndexOf(File.separator)));
    }

    /**
     * Returns project's faces. 
     * Note the they are not faces of individual analytical tasks.
     * 
     * @return project's faces
     */
    public FacesProxy getFacesProxy() {
        return faces;
    }
    
    /**
     * Returns project's name decoded from the name of the project's folder.
     * If the project is not yet saved then {@code DEFAULT_PROJECT_NAME} is used.
     * 
     * @return project's name
     */
    public String getProjectName() {
        if (hasProjectDir()) {
            String name = dir.toString().substring(
                dir.toString().lastIndexOf(File.separatorChar) + 1, 
                dir.toString().length()
            );
            if (name.endsWith(File.separator)) {
                name = name.substring(0, name.length()-1);
            }
            return name;
        } else {
            return DEFAULT_PROJECT_NAME;
        }
    }
    
    /**
     * Determines whether the project has changed and the should be saved.
     * @return {@code true} if the project has changed and the should be saved.
     */
    public boolean changed() {
        return changed;
    }
    
    /**
     * Determines whether the project has assigned a folder (has been saved in the past).
     * @return {@code true} if the project has assigned a folder (has been saved in the past)
     */
    public boolean hasProjectDir() {
        return dir != null;
    }
    
    /**
     * Returns project's folder or {@code null}
     * @return project's folder or {@code null}
     */
    public File getProjectDir() {
        return this.dir;
    }
    
    /**
     * Sets project's folder
     * @param path Path to the folder
     * @return {@code true} if the given path directs to existing folder
     */
    public boolean setProjectDir(File path) {
        if (path == null || !path.isDirectory()) {
            return false;
        } else {
            this.dir = path;
            return true;
        }
    }
    
    /**
     * Sets the project's modification indicator
     * @param changed modification indicator
     */
    public void setChanged(boolean changed) {
        this.changed = changed;
    }
    
    /**
     * Saves the project's state into the project's folder.
     * This method writes info about project's faces into the JSON file 
     * with the name {@code PROJECT_FILE_NAME} located in the project's folder.
     * This method doesn't write tasks on disk.
     * 
     * @return {@code true} if the project was successfully written
     * @throws IOException on IO errors
     */
    public boolean save() throws IOException {
        if (!hasProjectDir()) {
            return false;
        }
        
        File file = getProjectFile(dir);
        
        jacksonMapper.writeValue(file, new ProjectJsonStructure(faces));
        
        Preferences userPreferences = Preferences.userNodeForPackage(Project.class); 
        userPreferences.put(RECENT_PROJECT_PREF_KEY, file.getAbsolutePath());
        setChanged(false);
        
        return true;
    }
}
