package cz.fidentis.analyst.canvas;

import cz.fidentis.analyst.data.face.HumanFace;
import cz.fidentis.analyst.data.mesh.MeshTriangle;
import cz.fidentis.analyst.data.ray.Ray;
import cz.fidentis.analyst.data.ray.RayIntersection;
import cz.fidentis.analyst.drawables.Drawable;
import cz.fidentis.analyst.drawables.DrawableFace;
import cz.fidentis.analyst.engines.raycasting.RayIntersectionConfig;
import cz.fidentis.analyst.engines.raycasting.RayIntersectionServices;
import cz.fidentis.analyst.engines.face.FaceStateServices;
import cz.fidentis.analyst.engines.face.events.HumanFaceSelectedEvent;
import org.openide.util.Pair;

import javax.swing.*;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.Comparator;
import java.util.Objects;


/**
 * A mouse adapter that handles clicking in canvas and then casts the ray into
 * the scene. Moreover, the listener can be switched on/off.
 * 
 * @author Mario Chromik
 */
public class MouseClickListener extends MouseAdapter {
    private final Canvas canvas;
    private boolean active = true;
    
    /**
     * constructor requiring one parameter
     * @param canvas OpenGL canvas 
     */
    public MouseClickListener(Canvas canvas) {
        if (canvas == null) {
            throw new IllegalArgumentException("canvas is null");
        }
        this.canvas = canvas;
    }
    
    /**
     * toggles activity of object
     * @return current status
     */
    public boolean toggleActivity() {
        active = !active;
        return active;
    }
         
    /**
     * when mouse is clicked calculate a ray into the scene
     * @param evt mouse event
     */
    @Override
    public void mouseClicked(MouseEvent evt) {
        if (!SwingUtilities.isLeftMouseButton(evt) || !active){
            return;
        }

        Ray ray = canvas.getSceneRenderer().castRayThroughPixel(evt.getX(), evt.getY(), canvas.getCamera());

        Pair<HumanFace, RayIntersection> closestFace = canvas.getScene()
                .getDrawableFaces()
                .stream()
                .filter(Drawable::isShown)
                .map(DrawableFace::getHumanFace)
                .map(face -> getRayIntersection(face, ray))
                .filter(Objects::nonNull)
                .sorted(Comparator.comparingDouble(o -> o.second().getDistance()))
                .findFirst()
                .orElse(null);
        
        if (closestFace != null) {
            HumanFace face = closestFace.first();
            face.announceEvent(new HumanFaceSelectedEvent(
                    face, 
                    closestFace.second(), // distance
                    "Intersect with secondary face", 
                    this
            ));
        }                
    }
    
    protected Pair<HumanFace, RayIntersection> getRayIntersection(HumanFace face, Ray ray) {
        FaceStateServices.updateOctree(face, FaceStateServices.Mode.COMPUTE_IF_ABSENT);
        RayIntersection ri = RayIntersectionServices.computeClosest(
                face.getOctree(),
                new RayIntersectionConfig(ray, MeshTriangle.Smoothing.NORMAL, false));
        return (ri == null) ? null : Pair.of(face, ri);
    }
    
}
