package cz.fidentis.analyst.drawables;

import com.jogamp.opengl.GL2;
import com.jogamp.opengl.glu.GLU;
import com.jogamp.opengl.glu.GLUquadric;
import cz.fidentis.analyst.data.shapes.CrossSection3D;
import cz.fidentis.analyst.data.shapes.Plane;
import cz.fidentis.analyst.data.shapes.SurfaceMask;
import cz.fidentis.analyst.engines.cut.CrossSectionConfig;
import cz.fidentis.analyst.engines.cut.CrossSectionServices;

import javax.vecmath.Point3d;
import java.awt.*;
import java.util.ArrayList;
import java.util.List;


/**
 * Drawable interactive mask drawn as points connected by lines.
 * 
 * @author Mario Chromik
 */
public class DrawableInteractiveMask extends Drawable {
    
    private final SurfaceMask surfaceMask;
    
    private final double defaultRadius = 1;
    
    private final Color defaultColor = Color.WHITE;
    
    /**
     * Constructor.
     * 
     * @param mask interactive mask to generate drawable interactive mask from
     * @throws IllegalArgumentException if the {@code mask} is {@code null}
     */
    public DrawableInteractiveMask(SurfaceMask mask) {
        if (mask == null) {
            throw new IllegalArgumentException("mask");
        }
        surfaceMask = mask;
    }
    
    @Override
    protected void renderObject(GL2 gl) {
        float[] rgba = {
            defaultColor.getRed() / 255f, 
            defaultColor.getGreen() / 255f, 
            defaultColor.getBlue() / 255f, 
            defaultColor.getTransparency()};
            
        gl.glMaterialfv(GL2.GL_FRONT_AND_BACK, GL2.GL_DIFFUSE, rgba, 0);
        setPolygonMode(gl);
        
        // draw mask points
        for (SurfaceMask.MaskPoint maskPoint : surfaceMask.getPoints()) {
            gl.glPushMatrix(); 
            gl.glTranslated(maskPoint.getPosition().x, maskPoint.getPosition().y, maskPoint.getPosition().z);
            GLUquadric center = GLU_CONTEXT.gluNewQuadric();
            GLU_CONTEXT.gluQuadricDrawStyle(center, GLU.GLU_FILL);
            GLU_CONTEXT.gluQuadricNormals(center, GLU.GLU_FLAT);
            GLU_CONTEXT.gluQuadricOrientation(center, GLU.GLU_OUTSIDE);
            GLU_CONTEXT.gluSphere(center, defaultRadius, 16, 16);
            GLU_CONTEXT.gluDeleteQuadric(center);
            gl.glPopMatrix();
        }
        
        // draw interpolation curves between points:
        gl.glBegin(GL2.GL_LINES);
        for (CrossSection3D curve : interpolatePoints()) {
            for (List<Point3d> segment : curve.getSegments()) {
                for (int j = 0; j < segment.size() - 1; j++ ) {
                    Point3d p1 = segment.get(j);
                    Point3d p2 = segment.get(j + 1);
                    gl.glVertex3d(segment.get(j).x, segment.get(j).y, segment.get(j).z);
                    gl.glVertex3d(segment.get(j + 1).x, segment.get(j + 1).y, segment.get(j + 1).z); 
                }
            }
        }
        gl.glEnd();
    }
        
    protected List<CrossSection3D> interpolatePoints() {
        List<CrossSection3D> curves = new ArrayList<>();

        for (int i = 0; i < surfaceMask.getPoints().size() - 2; i++) {
            var p1 = surfaceMask.getPoints().get(i);
            var p2 = surfaceMask.getPoints().get(i + 1);
            var p3 = surfaceMask.getPoints().get(i + 2);
            Plane plane = new Plane(p1.getPosition(), p2.getPosition(), p3.getPosition());
            curves.add(CrossSectionServices.compute(p1.getFacet(), new CrossSectionConfig(plane)));
        }
        
        return curves;
    }
    
}
