package cz.fidentis.analyst.drawables;

import com.jogamp.opengl.GL2;
import com.jogamp.opengl.util.texture.Texture;
import com.jogamp.opengl.util.texture.TextureIO;
import cz.fidentis.analyst.data.mesh.*;
import cz.fidentis.analyst.rendering.RenderingMode;

import java.awt.*;
import java.io.File;
import java.util.List;

import static com.jogamp.opengl.GL.*;
import static com.jogamp.opengl.GL2ES1.GL_TEXTURE_ENV;
import static com.jogamp.opengl.GL2ES1.GL_TEXTURE_ENV_MODE;
import static com.jogamp.opengl.fixedfunc.GLLightingFunc.GL_SHININESS;
import static com.jogamp.opengl.fixedfunc.GLLightingFunc.GL_SPECULAR;

/**
 * A drawable triangular mesh, i.e., a mesh model with drawing information like 
 * material, transparency, color, relative transformations in the scene etc. 
 * This class encapsulates rendering state and parameters,
 * 
 * @author Radek Oslejsek
 * @author Richard Pajersky
 * @author Katerina Zarska
 */
public class DrawableMesh extends Drawable {
    
    private final MeshModel model;
    private Texture texture = null;
    
    /**
     * Copy constructor. 
     * 
     * @param mesh Mesh to be copied
     * @throws NullPointerException if the input argument is {@code null}
     */
    public DrawableMesh(DrawableMesh mesh) {
        super(mesh);
        this.model = MeshFactory.cloneMeshModel(mesh.getModel());
        this.texture = mesh.texture;
    }
    
    /**
     * Constructor. 
     * 
     * @param model Drawable mesh model
     * @throws IllegalArgumentException if the model is {@code null}
     */
    public DrawableMesh(MeshModel model) {
        if (model == null) {
            throw new IllegalArgumentException("model is null");
        }
        this.model = model;
    }
    
    /**
     * Constructor. 
     * 
     * @param facet Mesh facet
     * @throws IllegalArgumentException if the model is {@code null}
     */
    public DrawableMesh(MeshFacet facet) {
        if (facet == null) {
            throw new IllegalArgumentException("facet is null");
        }
        this.model = MeshFactory.createEmptyMeshModel();
        this.model.addFacet(facet);
    }
    
    @Override
    protected void initRendering(GL2 gl) {
        super.initRendering(gl);
        
        // set color of highlights
        Color col = getHighlights();
        float[] highlights = {col.getRed(), col.getGreen(), col.getBlue(), 1};
        gl.glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 50);
        gl.glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, highlights, 0);
        
        if (getRenderMode() == RenderingMode.TEXTURE && 
                model.hasMaterial() && 
                model.getMaterial().hasTexture() && 
                (texture == null)) {

            gl.glActiveTexture(GL_TEXTURE0);
            gl.glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
            gl.glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
            gl.glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
            try{
                texture = TextureIO.newTexture(new File(model.getMaterial().texturePath()), true);
                texture.bind(gl);
            } catch (Exception ex){
                ex.printStackTrace();
            }
        }
 
   }
    
    @Override
    protected void renderObject(GL2 gl) {
        if (texture != null && getRenderMode() == RenderingMode.TEXTURE) { // turn on temporarily
            gl.glActiveTexture(GL_TEXTURE0);
            gl.glBindTexture(GL_TEXTURE_2D, texture.getTextureObject());
            gl.glEnable(GL_TEXTURE_2D);
        }

        for (MeshFacet facet: getFacets()) {
            gl.glBegin(GL_TRIANGLES); //vertices are rendered as triangles

            for (MeshTriangle tri: facet) {
                for (MeshPoint mPoint :tri) {
                    // render the normals
                    if (mPoint.getNormal() != null) {
                        gl.glNormal3d(mPoint.getNormal().x, mPoint.getNormal().y, mPoint.getNormal().z);
                    }

                    // render the vertices
                    if (getRenderMode() == RenderingMode.TEXTURE && mPoint.getTexCoord() != null) {
                        gl.glTexCoord2d(mPoint.getTexCoord().x, mPoint.getTexCoord().y);
                    }

                    gl.glVertex3d(mPoint.getPosition().x, mPoint.getPosition().y, mPoint.getPosition().z);
                }
            }

            /*
            // get the normal and tex coordinate indices for face i
            for (int v = 0; v < facet.getCornerTable().getSize(); v++) {            
                MeshPoint mPoint = facet.getVertices().get(facet.getCornerTable().getRow(v).getVertexIndex());

                // render the normals
                if (mPoint.getNormal() != null) {
                    gl.glNormal3d(mPoint.getNormal().x, mPoint.getNormal().y, mPoint.getNormal().z);
                }

                // render the vertices
                if (getRenderMode() == RenderingMode.TEXTURE && mPoint.getTexCoord() != null) {
                    gl.glTexCoord2d(mPoint.getTexCoord().x, mPoint.getTexCoord().y);
                }
                gl.glVertex3d(mPoint.getPosition().x, mPoint.getPosition().y, mPoint.getPosition().z);
            }
             */
        
            gl.glEnd();
        }
        
        gl.glDisable(GL_TEXTURE_2D); // turn off
    }
    
    /**
     * Returns list of individual facets.
     * 
     * @return list of individual facets.
     */
    public List<MeshFacet> getFacets() {
        return model.getFacets();
    }
    
    /**
     * @return {@link MeshModel}
     */
    public MeshModel getModel() {
        return this.model;
    }

}
