package cz.fidentis.analyst.toolbar;

import cz.fidentis.analyst.canvas.Canvas;
import cz.fidentis.analyst.data.face.HumanFaceEvent;
import cz.fidentis.analyst.data.face.HumanFaceListener;
import cz.fidentis.analyst.drawables.DrawableFace;
import cz.fidentis.analyst.project.FacesProxy;
import cz.fidentis.analyst.rendering.ShadersManager;
import org.openide.awt.DropDownButtonFactory;
import org.openide.util.NbBundle;

import javax.swing.*;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.plaf.metal.MetalToggleButtonUI;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.image.BufferedImage;

/**
 * Canvas toolbox for scenes with two faces. 
 * It enables to show-hide faces in the scene, show-hide feature points,
 * and to change the transparency of face couple.
 * 
 * @author Richard Pajersky
 * @author Radek Oslejsek
 * @author Ondrej Simecek
 */
public class SceneToolboxFaceToFace extends JPanel implements HumanFaceListener {
    
    private static final String PRIMARY_FACE_ICON = "primaryFace28x28.png";
    private static final String SECONDARY_FACE_ICON = "secondaryFace28x28.png";
    private static final String FEATURE_POINTS_ICON = "fps28x28.png";
    private static final String FOG_VERSION_TOOLBOX_ICON = "fog_toolbox28x28.png";
    private static final String CONTOURS_ICON = "intersection28x28.png";
    private static final String GLYPH_ICON = "glyph28x28.png";

    public static final int DEFAULT_TRANSPARENCY = 75; // 0 = invisible, 100 = opaque
    
    private JToggleButton priLandButton;
    private JToggleButton secLandButton;
    private JToggleButton contoursButton;
    private JToggleButton glyphsButton;
    private JSlider       priSlider;
    private JSlider       secSlider;
    private JToolBar toolbar;
    
    private final Canvas canvas;
    
    /**
     * Faces processed by current analytical task
     */
    private final transient FacesProxy faces;
    
    /**
     * Constructor.
     * @param canvas Rendering canvas
     * @param faces Faces processed by current analytical task
     */
    public SceneToolboxFaceToFace(Canvas canvas, FacesProxy faces) {
        this.canvas = canvas;
        this.faces = faces;
        initComponents();  // buttons are turned on by default
        
        // be informed if something changes in the faces -- see acceptEvent()
        faces.getPrimaryFace().registerListener(this);
        faces.getSecondaryFace().registerListener(this);
        
        // Change initial state:

        // currently, FPs (landmarks) are either presented or not
        priLandButton.setSelected(false);
        if (faces.getPrimaryFace().hasLandmarks()) {
            canvas.getScene().getDrawableFeaturePoints(canvas.getScene().getPrimaryFaceSlot()).show(false);
        } else {
            priLandButton.setEnabled(false);
        }
        secLandButton.setSelected(false);
        if (faces.getSecondaryFace().hasLandmarks()) {
            canvas.getScene().getDrawableFeaturePoints(canvas.getScene().getSecondaryFaceSlot()).show(false);
        } else {
            secLandButton.setEnabled(false);
        }

        contoursButton.setSelected(false);
        glyphsButton.setSelected(false);

        priSlider.setValue(DEFAULT_TRANSPARENCY);
        canvas.getScene().getDrawableFace(
                canvas.getScene().getPrimaryFaceSlot()).setTransparency(DEFAULT_TRANSPARENCY/100f);
        secSlider.setValue(DEFAULT_TRANSPARENCY);
        canvas.getScene().getDrawableFace(
                canvas.getScene().getSecondaryFaceSlot()).setTransparency(DEFAULT_TRANSPARENCY/100f);
    }
    
    
    @Override
    public void acceptEvent(HumanFaceEvent event) {
    }
    
    private JToggleButton initButton(Color color, String icon, String tooltip) {
        JToggleButton button = new JToggleButton();
        button.setBorder(Canvas.BUTTONS_BORDER);
        if (color != null) {
            button.setBackground(null); // default color
            button.setUI(new MetalToggleButtonUI() {
                @Override
                protected Color getSelectColor() {
                    return color;
                }
            });
        }
        button.setIcon(new ImageIcon(getClass().getResource("/" + icon)));
        button.setFocusable(false);
        button.setSelected(true);
        button.setToolTipText(NbBundle.getMessage(SceneToolboxFaceToFace.class, tooltip));
        
        return button;
    }
    
    private JButton initColorButton(boolean primary, int width, int height) {
        JButton button = new JButton();
        button.setToolTipText("Color of the " + ((primary) ? "primary" : "secondary") + " face");
        button.setBorder(new EmptyBorder(0,0,0,0));
        BufferedImage image = new BufferedImage(width, height, java.awt.image.BufferedImage.TYPE_INT_RGB);
        Graphics2D graphics = image.createGraphics();
        graphics.setColor((primary) ? DrawableFace.SKIN_COLOR_PRIMARY : DrawableFace.SKIN_COLOR_SECONDARY);
        graphics.fillRect(0, 0, width, height);
        graphics.setXORMode(Color.LIGHT_GRAY);
        graphics.drawRect(0, 0, width - 1, height - 1);
        image.flush();
        button.setIcon(new ImageIcon(image));
        button.repaint();
        return button;
    }

    private void addToolbox(JPopupMenu toolbox, String icon, String tooltip) {
        // The button that will display the default action and the
        // dropdown arrow
        JButton button = DropDownButtonFactory.createDropDownButton(
                new ImageIcon(new BufferedImage(16, 16, BufferedImage.TYPE_INT_ARGB)),
                toolbox);

        button.setBorder(new EmptyBorder(5,5,5,5));
        button.setIcon(new ImageIcon(getClass().getResource("/" + icon)));

        button.setToolTipText(NbBundle.getMessage(Canvas.class, tooltip));
        button.setFocusable(false);
        toolbar.add(button);
    }

    private JSlider initSlider() {
        JSlider slider = new JSlider();
        slider.setPreferredSize(new Dimension(SceneToolboxSingleFace.DEFAULT_TRANSPARENCY_SLIDER_LENGTH, 50));
        slider.setMajorTickSpacing(50);
        slider.setMaximum(100);
        slider.setMinorTickSpacing(50);
        slider.setPaintTicks(false);
        slider.setToolTipText(org.openide.util.NbBundle.getMessage(SceneToolboxFaceToFace.class, "SceneToolboxFaceToFace.slider.tooltip"));
        slider.setValue(50);
        slider.setCursor(new java.awt.Cursor(java.awt.Cursor.HAND_CURSOR));
        slider.setBorder(new EmptyBorder(0,0,0,0));

        return slider;
    }

    private void initComponents() {
        JPanel panel = new JPanel();
        panel.setLayout(new FlowLayout());
        add(panel);

        toolbar = new JToolBar();
        toolbar.setFloatable(false);
        add(toolbar, BorderLayout.SOUTH);

        addToolbox(new FogVersionToolbox(canvas), FOG_VERSION_TOOLBOX_ICON, "SceneToolboxFaceToFace.fog.tooltip");

        priLandButton = initButton(null, FEATURE_POINTS_ICON, "SceneToolboxFaceToFace.landButton.tooltip");
        panel.add(priLandButton);
        priLandButton.addActionListener(new AbstractAction() {
            @Override
            public void actionPerformed(ActionEvent e) {
                boolean onOff = ((JToggleButton) e.getSource()).isSelected();
                canvas.getScene().getDrawableFeaturePoints(canvas.getScene().getPrimaryFaceSlot()).show(onOff);
                canvas.renderScene();
            }
        });

        JToggleButton priFaceButton = initButton(null, PRIMARY_FACE_ICON, "SceneToolboxFaceToFace.faceButton.tooltip");
        panel.add(priFaceButton);
        priFaceButton.addActionListener(new AbstractAction() {
            @Override
            public void actionPerformed(ActionEvent e) {
                boolean onOff = ((JToggleButton) e.getSource()).isSelected();
                canvas.getScene().showDrawableFace(canvas.getScene().getPrimaryFaceSlot(), onOff);
                canvas.renderScene();
            }
        });
        
        /////////////////////////
        JButton priColorButton = initColorButton(true, 16, 38);
        panel.add(priColorButton);
        
        /////////////////////////
        priSlider = initSlider();
        priSlider.addChangeListener((ChangeEvent e) -> {
            int val = ((JSlider) e.getSource()).getValue();
            canvas.getScene().getDrawableFace(canvas.getScene().getPrimaryFaceSlot()).setTransparency(val/100f);
            canvas.renderScene();
        });
        panel.add(priSlider);

        secLandButton = initButton(null, FEATURE_POINTS_ICON, "SceneToolboxFaceToFace.landButton.tooltip");
        panel.add(secLandButton);
        secLandButton.addActionListener(new AbstractAction() {
            @Override
            public void actionPerformed(ActionEvent e) {
                boolean onOff = ((JToggleButton) e.getSource()).isSelected();
                canvas.getScene().getDrawableFeaturePoints(canvas.getScene().getSecondaryFaceSlot()).show(onOff);
                canvas.renderScene();
            }
        });

        /////////////////////////
        JToggleButton secFaceButton = initButton(null, SECONDARY_FACE_ICON, "SceneToolboxFaceToFace.faceButton.tooltip");
        panel.add(secFaceButton);
        secFaceButton.addActionListener(new AbstractAction() {
            @Override
            public void actionPerformed(ActionEvent e) {
                boolean onOff = ((JToggleButton) e.getSource()).isSelected();
                canvas.getScene().showDrawableFace(canvas.getScene().getSecondaryFaceSlot(), onOff);
                canvas.renderScene();
            }
        });

        /////////////////////////
        JButton secColorButton = initColorButton(false, 16, 38);
        panel.add(secColorButton);

        /////////////////////////
        secSlider = initSlider();
        secSlider.addChangeListener((ChangeEvent e) -> {
            int val = ((JSlider) e.getSource()).getValue();
            canvas.getScene().getDrawableFace(canvas.getScene().getSecondaryFaceSlot()).setTransparency(val/100f);
            canvas.renderScene();
        });
        panel.add(secSlider);

        contoursButton = initButton(null, CONTOURS_ICON, "SceneToolboxFaceToFace.contoursButton.tooltip");
        panel.add(contoursButton);
        contoursButton.addActionListener(new AbstractAction() {
            @Override
            public void actionPerformed(ActionEvent e) {
                boolean onOff = ((JToggleButton) e.getSource()).isSelected();
                if (canvas.getSceneRenderer().getShadersManager() != null) {
                    canvas.getSceneRenderer().getShadersManager().setUseContours(onOff);
                }
                canvas.renderScene();
            }
        });

        glyphsButton = initButton(null, GLYPH_ICON, "SceneToolboxFaceToFace.contoursButton.tooltip");
        panel.add(glyphsButton);
        glyphsButton.addActionListener(new AbstractAction() {
            @Override
            public void actionPerformed(ActionEvent e) {
                boolean onOff = ((JToggleButton) e.getSource()).isSelected();
                ShadersManager shaders = canvas.getSceneRenderer().getShadersManager();
                if (shaders != null) {
                    shaders.setUseGlyphs(onOff);
                    canvas.renderScene();
                }
            }
        });
    }
}
