package cz.fidentis.analyst.toolbar;

import cz.fidentis.analyst.canvas.Canvas;
import cz.fidentis.analyst.data.face.HumanFaceEvent;
import cz.fidentis.analyst.data.face.HumanFaceListener;
import cz.fidentis.analyst.project.FacesProxy;
import org.openide.util.NbBundle;

import javax.swing.*;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.plaf.metal.MetalToggleButtonUI;
import java.awt.*;
import java.awt.event.ActionEvent;

/**
 * Canvas toolbox for scenes with a single faces. 
 * It enables to show-hide the face and its feature points,
 * and to change the transparency of the face.
 * 
 * @author Richard Pajersky
 * @author Radek Oslejsek
 */
public class SceneToolboxSingleFace extends JPanel implements HumanFaceListener {
    
    private static final String FACE_ICON = "head28x28.png";
    private static final String FEATURE_POINTS_ICON = "fps28x28.png";
    
    public static final int DEFAULT_TRANSPARENCY = 100; // 0 = invisible, 100 = opaque
    public static final int DEFAULT_TRANSPARENCY_SLIDER_LENGTH = 150;

    private JToggleButton priLandButton;
    private JToggleButton priFaceButton;
    private JSlider       slider;
    
    private final Canvas canvas;
    
    /**
     * Faces processed by current analytical task
     */
    private FacesProxy faces;
    
    /**
     * Constructor.
     * @param canvas Rendering canvas
     * @param faces Faces processed by current analytical task
     */
    public SceneToolboxSingleFace(Canvas canvas, FacesProxy faces) {
        this.canvas = canvas;
        this.faces = faces;
        initComponents();
        
        // currently, FPs (landmarks) are either presented or not
        priLandButton.setSelected(false);
        if (faces.getPrimaryFace() != null) {
            if (faces.getPrimaryFace().hasLandmarks() && canvas.getScene().getPrimaryFaceSlot() != -1) {
                canvas.getScene().getDrawableFeaturePoints(canvas.getScene().getPrimaryFaceSlot()).show(false);
            } else {
                priLandButton.setEnabled(false);
            }
        
            // be informed if something changes in the faces -- see acceptEvent()
            faces.getPrimaryFace().registerListener(this);
        }
    }
    
    @Override
    public void acceptEvent(HumanFaceEvent event) {
    }

    private JToggleButton initButton(Color color, String icon, String tooltip) {
        JToggleButton button = new JToggleButton();
        button.setBorder(Canvas.BUTTONS_BORDER);
        if (color != null) {
            button.setBackground(null); // default color
            button.setUI(new MetalToggleButtonUI() {
                @Override
                protected Color getSelectColor() {
                    return color;
                }
            });
        }
        button.setIcon(new ImageIcon(getClass().getResource("/" + icon)));
        button.setFocusable(false);
        button.setSelected(true);
        button.setToolTipText(NbBundle.getMessage(SceneToolboxFaceToFace.class, tooltip));
        
        return button;
    }

    private void initComponents() {
        JPanel panel = new JPanel();
        panel.setLayout(new FlowLayout());
        add(panel);
        
        priLandButton = initButton(null, FEATURE_POINTS_ICON, "SceneToolboxFaceToFace.landButton.tooltip");
        panel.add(priLandButton);
        priLandButton.addActionListener(new AbstractAction() {
            @Override
            public void actionPerformed(ActionEvent e) {
                boolean onOff = ((JToggleButton) e.getSource()).isSelected();
                canvas.getScene().getDrawableFeaturePoints(canvas.getScene().getPrimaryFaceSlot()).show(onOff);
                canvas.renderScene();
            }
        });
        
        priFaceButton = initButton(null, FACE_ICON, "SceneToolboxFaceToFace.faceButton.tooltip");
        panel.add(priFaceButton);
        priFaceButton.addActionListener(new AbstractAction() {
            @Override
            public void actionPerformed(ActionEvent e) {
                boolean onOff = ((JToggleButton) e.getSource()).isSelected();
                canvas.getScene().showDrawableFace(canvas.getScene().getPrimaryFaceSlot(), onOff);
                canvas.renderScene();
            }
        });
        
        /////////////////////////
        slider = new JSlider();
        slider.setPreferredSize(new Dimension(DEFAULT_TRANSPARENCY_SLIDER_LENGTH, 50));
        slider.setMaximum(100);
        slider.setPaintTicks(false);
        slider.setToolTipText(org.openide.util.NbBundle.getMessage(SceneToolboxSingleFace.class, "SceneToolboxSingleFace.slider.tooltip"));
        slider.setValue(DEFAULT_TRANSPARENCY);
        slider.setCursor(new java.awt.Cursor(java.awt.Cursor.HAND_CURSOR));
        slider.setBorder(new EmptyBorder(0,0,0,0));
        panel.add(slider);
        
        slider.addChangeListener((ChangeEvent e) -> {
            int val = ((JSlider) e.getSource()).getValue();
            canvas.getScene().getDrawableFace(canvas.getScene().getPrimaryFaceSlot()).setTransparency(val/100f);
            canvas.renderScene();
        });
    }
}
